@@ -26,7 +26,7 @@
 #ifndef _COMPIZ_H
 #define _COMPIZ_H
 
-#define ABIVERSION 20070213
+#define ABIVERSION 20070708
 
 #include <stdio.h>
 #include <sys/time.h>
@@ -40,6 +40,8 @@
 #include <X11/Xregion.h>
 #include <X11/XKBlib.h>
 
+#include <libxml/parser.h>
+
 #define SN_API_NOT_YET_FROZEN
 #include <libsn/sn.h>
 
@@ -69,10 +71,6 @@ extern "C" {
 #  define BITMAP_BIT_ORDER LSBFirst
 #endif
 
-/* For now, we never actually translate strings inside compiz */
-#define  _(x) x
-#define N_(x) x
-
 typedef struct _CompPlugin	  CompPlugin;
 typedef struct _CompDisplay	  CompDisplay;
 typedef struct _CompScreen	  CompScreen;
@@ -84,6 +82,11 @@ typedef struct _CompProgram	  CompProgram;
 typedef struct _CompFunction	  CompFunction;
 typedef struct _CompFunctionData  CompFunctionData;
 typedef struct _FragmentAttrib    FragmentAttrib;
+typedef struct _CompCursor	  CompCursor;
+typedef struct _CompMatch	  CompMatch;
+typedef struct _CompMetadata      CompMetadata;
+typedef struct _CompOutput        CompOutput;
+typedef struct _CompWalker        CompWalker;
 
 /* virtual modifiers */
 
@@ -156,6 +159,8 @@ typedef struct _FragmentAttrib    FragmentAttrib;
 #define CompWindowActionCloseMask	  (1 << 7)
 #define CompWindowActionShadeMask	  (1 << 8)
 #define CompWindowActionChangeDesktopMask (1 << 9)
+#define CompWindowActionAboveMask	  (1 << 10)
+#define CompWindowActionBelowMask	  (1 << 11)
 
 #define MwmFuncAll      (1L << 0)
 #define MwmFuncResize   (1L << 1)
@@ -206,6 +211,7 @@ extern REGION     emptyRegion;
 extern REGION     infiniteRegion;
 extern GLushort   defaultColor[4];
 extern Window     currentRoot;
+extern Bool       shutDown;
 extern Bool       restartSignal;
 extern CompWindow *lastFoundWindow;
 extern CompWindow *lastDamagedWindow;
@@ -214,18 +220,19 @@ extern Bool       indirectRendering;
 extern Bool       strictBinding;
 extern Bool       useCow;
 extern Bool       noDetection;
+extern Bool	  useDesktopHints;
+extern Bool       onlyCurrentScreen;
 
 extern int  defaultRefreshRate;
 extern char *defaultTextureFilter;
 
-extern char *windowTypeString[];
-extern int  nWindowTypeString;
-
 extern int lastPointerX;
 extern int lastPointerY;
 extern int pointerX;
 extern int pointerY;
 
+extern CompMetadata coreMetadata;
+
 #define RESTRICT_VALUE(value, min, max)				     \
     (((value) < (min)) ? (min): ((value) > (max)) ? (max) : (value))
 
@@ -271,6 +278,7 @@ typedef enum {
     CompOptionTypeString,
     CompOptionTypeColor,
     CompOptionTypeAction,
+    CompOptionTypeMatch,
     CompOptionTypeList
 } CompOptionType;
 
@@ -292,9 +300,19 @@ typedef enum {
     CompActionStateInitEdgeDnd = 1 <<  7,
     CompActionStateTermEdgeDnd = 1 <<  8,
     CompActionStateCommit      = 1 <<  9,
-    CompActionStateCancel      = 1 << 10
+    CompActionStateCancel      = 1 << 10,
+    CompActionStateAutoGrab    = 1 << 11
 } CompActionState;
 
+typedef enum {
+    CompLogLevelFatal = 0,
+    CompLogLevelError,
+    CompLogLevelWarn,
+    CompLogLevelInfo,
+    CompLogLevelDebug
+
+} CompLogLevel;
+
 typedef struct _CompKeyBinding {
     int		 keycode;
     unsigned int modifiers;
@@ -330,6 +348,16 @@ struct _CompAction {
 
     unsigned int edgeMask;
     int		 edgeButton;
+
+    CompPrivate priv;
+};
+
+typedef union _CompMatchOp CompMatchOp;
+
+struct _CompMatch {
+    CompDisplay *display;
+    CompMatchOp *op;
+    int		nOp;
 };
 
 typedef struct {
@@ -345,6 +373,7 @@ union _CompOptionValue {
     char	   *s;
     unsigned short c[4];
     CompAction     action;
+    CompMatch      match;
     CompListValue  list;
 };
 
@@ -359,21 +388,13 @@ typedef struct _CompOptionFloatRestriction {
     float precision;
 } CompOptionFloatRestriction;
 
-typedef struct _CompOptionStringRestriction {
-    char **string;
-    int  nString;
-} CompOptionStringRestriction;
-
 typedef union {
     CompOptionIntRestriction    i;
     CompOptionFloatRestriction  f;
-    CompOptionStringRestriction s;
 } CompOptionRestriction;
 
 struct _CompOption {
     char		  *name;
-    char		  *shortDesc;
-    char		  *longDesc;
     CompOptionType	  type;
     CompOptionValue	  value;
     CompOptionRestriction rest;
@@ -382,6 +403,12 @@ struct _CompOption {
 typedef CompOption *(*DisplayOptionsProc) (CompDisplay *display, int *count);
 typedef CompOption *(*ScreenOptionsProc) (CompScreen *screen, int *count);
 
+void
+compInitOption (CompOption *option);
+
+void
+compFiniOption (CompOption *option);
+
 CompOption *
 compFindOption (CompOption *option,
 		int	    nOption,
@@ -412,10 +439,18 @@ Bool
 compSetActionOption (CompOption      *option,
 		     CompOptionValue *value);
 
+Bool
+compSetMatchOption (CompOption      *option,
+		    CompOptionValue *value);
+
 Bool
 compSetOptionList (CompOption      *option,
 		   CompOptionValue *value);
 
+Bool
+compSetOption (CompOption      *option,
+	       CompOptionValue *value);
+
 unsigned int
 compWindowTypeMaskFromStringList (CompOptionValue *value);
 
@@ -449,6 +484,12 @@ getColorOptionNamed (CompOption	    *option,
 		     char	    *name,
 		     unsigned short *defaultValue);
 
+CompMatch *
+getMatchOptionNamed (CompOption	*option,
+		     int	nOption,
+		     char	*name,
+		     CompMatch  *defaultValue);
+
 char *
 keyBindingToString (CompDisplay    *d,
 		    CompKeyBinding *key);
@@ -550,9 +591,20 @@ typedef int CompFileWatchHandle;
 
 typedef CompOption *(*GetDisplayOptionsProc) (CompDisplay *display,
 					      int	  *count);
+
 typedef Bool (*SetDisplayOptionProc) (CompDisplay     *display,
 				      char	      *name,
 				      CompOptionValue *value);
+
+typedef CompOption *(*GetPluginDisplayOptionsProc) (CompPlugin  *plugin,
+						    CompDisplay *display,
+						    int	        *count);
+
+typedef Bool (*SetPluginDisplayOptionProc) (CompPlugin      *plugin,
+					    CompDisplay     *display,
+					    char	    *name,
+					    CompOptionValue *value);
+
 typedef Bool (*SetDisplayOptionForPluginProc) (CompDisplay     *display,
 					       char	       *plugin,
 					       char	       *name,
@@ -598,6 +650,7 @@ typedef Bool (*ImageToFileProc) (CompDisplay *display,
 #define NOTIFY_CREATE_MASK (1 << 0)
 #define NOTIFY_DELETE_MASK (1 << 1)
 #define NOTIFY_MOVE_MASK   (1 << 2)
+#define NOTIFY_MODIFY_MASK (1 << 3)
 
 typedef void (*FileWatchCallBackProc) (const char *name,
 				       void	  *closure);
@@ -617,6 +670,67 @@ typedef void (*FileWatchAddedProc) (CompDisplay	  *display,
 typedef void (*FileWatchRemovedProc) (CompDisplay   *display,
 				      CompFileWatch *fileWatch);
 
+#define MATCH_OP_AND_MASK (1 << 0)
+#define MATCH_OP_NOT_MASK (1 << 1)
+
+typedef enum {
+    CompMatchOpTypeGroup,
+    CompMatchOpTypeExp
+} CompMatchOpType;
+
+typedef struct _CompMatchAnyOp {
+    CompMatchOpType type;
+    int		    flags;
+} CompMatchAnyOp;
+
+typedef struct _CompMatchGroupOp {
+    CompMatchOpType type;
+    int		    flags;
+    CompMatchOp	    *op;
+    int		    nOp;
+} CompMatchGroupOp;
+
+typedef void (*CompMatchExpFiniProc) (CompDisplay *display,
+				      CompPrivate priv);
+
+typedef Bool (*CompMatchExpEvalProc) (CompDisplay *display,
+				      CompWindow  *window,
+				      CompPrivate priv);
+
+typedef struct _CompMatchExp {
+    CompMatchExpFiniProc fini;
+    CompMatchExpEvalProc eval;
+    CompPrivate		 priv;
+} CompMatchExp;
+
+typedef struct _CompMatchExpOp {
+    CompMatchOpType type;
+    int		    flags;
+    char	    *value;
+    CompMatchExp    e;
+} CompMatchExpOp;
+
+union _CompMatchOp {
+    CompMatchOpType  type;
+    CompMatchAnyOp   any;
+    CompMatchGroupOp group;
+    CompMatchExpOp   exp;
+};
+
+typedef void (*MatchInitExpProc) (CompDisplay  *display,
+				  CompMatchExp *exp,
+				  const char   *value);
+
+typedef void (*MatchExpHandlerChangedProc) (CompDisplay *display);
+
+typedef void (*MatchPropertyChangedProc) (CompDisplay *display,
+					  CompWindow  *window);
+
+typedef void (*LogMessageProc) (CompDisplay  *d,
+				char         *componentName,
+				CompLogLevel level,
+				char         *message);
+
 struct _CompDisplay {
     Display    *display;
     CompScreen *screens;
@@ -626,8 +740,11 @@ struct _CompDisplay {
 
     int compositeEvent, compositeError, compositeOpcode;
     int damageEvent, damageError;
-    int randrEvent, randrError;
     int syncEvent, syncError;
+    int fixesEvent, fixesError, fixesVersion;
+
+    Bool randrExtension;
+    int randrEvent, randrError;
 
     Bool shapeExtension;
     int  shapeEvent, shapeError;
@@ -704,6 +821,8 @@ struct _CompDisplay {
     Atom winActionCloseAtom;
     Atom winActionShadeAtom;
     Atom winActionChangeDesktopAtom;
+    Atom winActionAboveAtom;
+    Atom winActionBelowAtom;
 
     Atom wmAllowedActionsAtom;
 
@@ -807,9 +926,16 @@ struct _CompDisplay {
     FileWatchAddedProc   fileWatchAdded;
     FileWatchRemovedProc fileWatchRemoved;
 
+    MatchInitExpProc	       matchInitExp;
+    MatchExpHandlerChangedProc matchExpHandlerChanged;
+    MatchPropertyChangedProc   matchPropertyChanged;
+
+    LogMessageProc logMessage;
+
     CompPrivate *privates;
 };
 
+
 extern CompDisplay *compDisplays;
 
 int
@@ -827,7 +953,7 @@ compAddTimeout (int	     time,
 		CallBackProc callBack,
 		void	     *closure);
 
-void
+void *
 compRemoveTimeout (CompTimeoutHandle handle);
 
 CompWatchFdHandle
@@ -839,6 +965,9 @@ compAddWatchFd (int	     fd,
 void
 compRemoveWatchFd (CompWatchFdHandle handle);
 
+short int
+compWatchFdEvents (CompWatchFdHandle handle);
+
 CompFileWatchHandle
 addFileWatch (CompDisplay	    *display,
 	      const char	    *path,
@@ -858,6 +987,22 @@ void
 fileWatchRemoved (CompDisplay   *display,
 		  CompFileWatch *fileWatch);
 
+void
+compLogMessage (CompDisplay  *d,
+		char         *componentName,
+		CompLogLevel level,
+		char         *format,
+		...);
+
+void
+logMessage (CompDisplay  *d,
+	    char         *componentName,
+	    CompLogLevel level,
+	    char         *message);
+
+const char *
+logLevelToString (CompLogLevel level);
+
 int
 compCheckForError (Display *dpy);
 
@@ -866,9 +1011,7 @@ addScreenToDisplay (CompDisplay *display,
 		    CompScreen *s);
 
 Bool
-addDisplay (char *name,
-	    char **plugin,
-	    int  nPlugin);
+addDisplay (char *name);
 
 Time
 getCurrentTimeFromDisplay (CompDisplay *d);
@@ -916,7 +1059,7 @@ handleSelectionClear (CompDisplay *display,
 		      XEvent      *event);
 
 void
-warpPointer (CompDisplay *display,
+warpPointer (CompScreen *screen,
 	     int	 dx,
 	     int	 dy);
 
@@ -960,6 +1103,9 @@ imageToFile (CompDisplay *display,
 	     int	 stride,
 	     void	 *data);
 
+CompCursor *
+findCursorAtDisplay (CompDisplay *display);
+
 
 /* event.c */
 
@@ -1054,26 +1200,47 @@ typedef void (*DonePaintScreenProc) (CompScreen *screen);
 #define PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK (1 << 3)
 #define PAINT_SCREEN_CLEAR_MASK			   (1 << 4)
 
-typedef Bool (*PaintScreenProc) (CompScreen		 *screen,
+typedef void (*PaintScreenProc) (CompScreen   *screen,
+				 CompOutput   *outputs,
+				 int          numOutput,
+				 unsigned int mask);
+
+typedef Bool (*PaintOutputProc) (CompScreen		 *screen,
 				 const ScreenPaintAttrib *sAttrib,
 				 const CompTransform	 *transform,
 				 Region			 region,
-				 int		         output,
+				 CompOutput		 *output,
 				 unsigned int		 mask);
 
-typedef void (*PaintTransformedScreenProc) (CompScreen		    *screen,
+typedef void (*PaintTransformedOutputProc) (CompScreen		    *screen,
 					    const ScreenPaintAttrib *sAttrib,
 					    const CompTransform	    *transform,
 					    Region		    region,
-					    int			    output,
+					    CompOutput		    *output,
 					    unsigned int	    mask);
 
 /* XXX: ApplyScreenTransformProc will be removed */
 typedef void (*ApplyScreenTransformProc) (CompScreen		  *screen,
 					  const ScreenPaintAttrib *sAttrib,
-					  int			  output,
+					  CompOutput		  *output,
 					  CompTransform	          *transform);
 
+typedef void (*WalkerFiniProc) (CompScreen *screen,
+				CompWalker *walker);
+
+typedef CompWindow *(*WalkInitProc) (CompScreen *screen);
+typedef CompWindow *(*WalkStepProc) (CompWindow *window);
+
+struct _CompWalker {
+    WalkerFiniProc fini;
+    CompPrivate	   priv;
+
+    WalkInitProc first;
+    WalkInitProc last;
+    WalkStepProc next;
+    WalkStepProc prev;
+};
+
 /*
   window paint flags
 
@@ -1085,23 +1252,17 @@ typedef void (*ApplyScreenTransformProc) (CompScreen		  *screen,
   that is used to optimize rendering.
 */
 
-/*
-  this flag is present when we are painting clipped
-  opaque window regions only.
-*/
-#define PAINT_WINDOW_CLIP_OPAQUE_MASK		(1 << 0)
-
-/*
-  this flag is present when we are painting clipped
-  translucent window regions only.
-*/
-#define PAINT_WINDOW_CLIP_TRANSLUCENT_MASK	(1 << 1)
-
 /*
   this flag is present when window is being painted
   on a transformed screen.
 */
-#define PAINT_WINDOW_ON_TRANSFORMED_SCREEN_MASK (1 << 2)
+#define PAINT_WINDOW_ON_TRANSFORMED_SCREEN_MASK (1 << 0)
+
+/*
+  this flag is present when window is being tested
+  for occlusion of other windows.
+*/
+#define PAINT_WINDOW_OCCLUSION_DETECTION_MASK   (1 << 1)
 
 /*
   flag indicate that window is translucent.
@@ -1150,6 +1311,11 @@ typedef void (*DrawWindowTextureProc) (CompWindow	    *w,
 
 typedef void (*DrawWindowGeometryProc) (CompWindow *window);
 
+typedef void (*PaintCursorProc) (CompCursor	     *cursor,
+				 const CompTransform *transform,
+				 Region		     region,
+				 unsigned int	     mask);
+
 #define PAINT_BACKGROUND_ON_TRANSFORMED_SCREEN_MASK (1 << 0)
 
 typedef void (*PaintBackgroundProc) (CompScreen   *screen,
@@ -1165,37 +1331,43 @@ donePaintScreen (CompScreen *screen);
 
 void
 transformToScreenSpace (CompScreen    *screen,
-			int	      output,
+			CompOutput    *output,
 			float         z,
 			CompTransform *transform);
 
 /* XXX: prepareXCoords will be removed */
 void
 prepareXCoords (CompScreen *screen,
-		int	   output,
+		CompOutput *output,
 		float      z);
 
 void
-paintTransformedScreen (CompScreen		*screen,
+paintTransformedOutput (CompScreen		*screen,
 			const ScreenPaintAttrib *sAttrib,
 			const CompTransform	*transform,
 			Region			region,
-			int			output,
+			CompOutput		*output,
 			unsigned int	        mask);
 
 /* XXX: applyScreenTransform will be removed */
 void
 applyScreenTransform (CompScreen	      *screen,
 		      const ScreenPaintAttrib *sAttrib,
-		      int		      output,
+		      CompOutput	      *output,
 		      CompTransform	      *transform);
 
+void
+paintScreen (CompScreen   *screen,
+	     CompOutput   *outputs,
+	     int          numOutput,
+	     unsigned int mask);
+
 Bool
-paintScreen (CompScreen		     *screen,
+paintOutput (CompScreen		     *screen,
 	     const ScreenPaintAttrib *sAttrib,
 	     const CompTransform     *transform,
 	     Region		     region,
-	     int		     output,
+	     CompOutput		     *output,
 	     unsigned int	     mask);
 
 Bool
@@ -1233,6 +1405,12 @@ paintWindow (CompWindow		     *w,
 	     Region		     region,
 	     unsigned int	     mask);
 
+void
+paintCursor (CompCursor		 *cursor,
+	     const CompTransform *transform,
+	     Region		 region,
+	     unsigned int	 mask);
+
 void
 paintBackground (CompScreen   *screen,
 		 Region	      region,
@@ -1276,6 +1454,23 @@ void
 destroyTexture (CompScreen  *screen,
 		CompTexture *texture);
 
+Bool
+imageBufferToTexture (CompScreen   *screen,
+		      CompTexture  *texture,
+		      char	   *image,
+		      unsigned int width,
+		      unsigned int height);
+
+Bool
+imageDataToTexture (CompScreen   *screen,
+		    CompTexture  *texture,
+		    char	 *image,
+		    unsigned int width,
+		    unsigned int height,
+		    GLenum       format,
+		    GLenum       type);
+
+
 Bool
 readImageToTexture (CompScreen   *screen,
 		    CompTexture  *texture,
@@ -1321,20 +1516,22 @@ disableTexture (CompScreen  *screen,
 
 /* screen.c */
 
-#define COMP_SCREEN_OPTION_DETECT_REFRESH_RATE 0
-#define COMP_SCREEN_OPTION_LIGHTING	       1
-#define COMP_SCREEN_OPTION_REFRESH_RATE	       2
-#define COMP_SCREEN_OPTION_HSIZE	       3
-#define COMP_SCREEN_OPTION_VSIZE	       4
-#define COMP_SCREEN_OPTION_OPACITY_STEP        5
-#define COMP_SCREEN_OPTION_UNREDIRECT_FS       6
-#define COMP_SCREEN_OPTION_DEFAULT_ICON        7
-#define COMP_SCREEN_OPTION_SYNC_TO_VBLANK      8
-#define COMP_SCREEN_OPTION_NUMBER_OF_DESKTOPS  9
-#define COMP_SCREEN_OPTION_DETECT_OUTPUTS      10
-#define COMP_SCREEN_OPTION_OUTPUTS	       11
-#define COMP_SCREEN_OPTION_FOCUS_PREVENTION    12
-#define COMP_SCREEN_OPTION_NUM		       13
+#define COMP_SCREEN_OPTION_DETECT_REFRESH_RATE	  0
+#define COMP_SCREEN_OPTION_LIGHTING		  1
+#define COMP_SCREEN_OPTION_REFRESH_RATE		  2
+#define COMP_SCREEN_OPTION_HSIZE		  3
+#define COMP_SCREEN_OPTION_VSIZE		  4
+#define COMP_SCREEN_OPTION_OPACITY_STEP		  5
+#define COMP_SCREEN_OPTION_UNREDIRECT_FS	  6
+#define COMP_SCREEN_OPTION_DEFAULT_ICON		  7
+#define COMP_SCREEN_OPTION_SYNC_TO_VBLANK	  8
+#define COMP_SCREEN_OPTION_NUMBER_OF_DESKTOPS	  9
+#define COMP_SCREEN_OPTION_DETECT_OUTPUTS	  10
+#define COMP_SCREEN_OPTION_OUTPUTS		  11
+#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_MATCH 12
+#define COMP_SCREEN_OPTION_OPACITY_MATCHES	  13
+#define COMP_SCREEN_OPTION_OPACITY_VALUES	  14
+#define COMP_SCREEN_OPTION_NUM		          15
 
 #ifndef GLX_EXT_texture_from_pixmap
 #define GLX_BIND_TO_TEXTURE_RGB_EXT        0x20D0
@@ -1441,6 +1638,13 @@ typedef CompOption *(*GetScreenOptionsProc) (CompScreen *screen,
 typedef Bool (*SetScreenOptionProc) (CompScreen      *screen,
 				     char	     *name,
 				     CompOptionValue *value);
+typedef CompOption *(*GetPluginScreenOptionsProc) (CompPlugin *plugin,
+						   CompScreen *screen,
+						   int	      *count);
+typedef Bool (*SetPluginScreenOptionProc) (CompPlugin      *plugin,
+					   CompScreen      *screen,
+					   char	           *name,
+					   CompOptionValue *value);
 typedef Bool (*SetScreenOptionForPluginProc) (CompScreen      *screen,
 					      char	      *plugin,
 					      char	      *name,
@@ -1452,6 +1656,11 @@ typedef Bool (*InitPluginForScreenProc) (CompPlugin *plugin,
 typedef void (*FiniPluginForScreenProc) (CompPlugin *plugin,
 					 CompScreen *screen);
 
+typedef void (*EnterShowDesktopModeProc) (CompScreen *screen);
+
+typedef void (*LeaveShowDesktopModeProc) (CompScreen *screen,
+					  CompWindow *window);
+
 typedef Bool (*DamageWindowRectProc) (CompWindow *w,
 				      Bool       initial,
 				      BoxPtr     rect);
@@ -1459,12 +1668,29 @@ typedef Bool (*DamageWindowRectProc) (CompWindow *w,
 typedef Bool (*DamageWindowRegionProc) (CompWindow *w,
 					Region     region);
 
+typedef Bool (*DamageCursorRectProc) (CompCursor *c,
+				      Bool       initial,
+				      BoxPtr     rect);
+
+
 typedef void (*GetOutputExtentsForWindowProc) (CompWindow	 *w,
 					       CompWindowExtents *output);
 
+typedef unsigned int (*GetAllowedActionsForWindowProc) (CompWindow *w);
+
 typedef Bool (*FocusWindowProc) (CompWindow *window);
 
-typedef void (*WindowResizeNotifyProc) (CompWindow *window);
+typedef Bool (*PlaceWindowProc) (CompWindow *window,
+				 int        x,
+				 int        y,
+				 int        *newX,
+				 int        *newY);
+
+typedef void (*WindowResizeNotifyProc) (CompWindow *window,
+					int        dx,
+					int        dy,
+					int        dwidth,
+					int        dheight);
 
 typedef void (*WindowMoveNotifyProc) (CompWindow *window,
 				      int	 dx,
@@ -1486,8 +1712,13 @@ typedef void (*WindowUngrabNotifyProc) (CompWindow *window);
 
 typedef void (*WindowStateChangeNotifyProc) (CompWindow *window);
 
+typedef void (*WindowAddNotifyProc) (CompWindow *window);
+
 typedef void (*OutputChangeNotifyProc) (CompScreen *screen);
 
+typedef void (*InitWindowWalkerProc) (CompScreen *screen,
+				      CompWalker *walker);
+
 #define COMP_SCREEN_DAMAGE_PENDING_MASK (1 << 0)
 #define COMP_SCREEN_DAMAGE_REGION_MASK  (1 << 1)
 #define COMP_SCREEN_DAMAGE_ALL_MASK     (1 << 2)
@@ -1528,6 +1759,7 @@ typedef struct _CompFBConfig {
     int         yInverted;
     int         mipmap;
     int         textureFormat;
+    int         textureTargets;
 } CompFBConfig;
 
 #define NOTHING_TRANS_FILTER 0
@@ -1555,13 +1787,48 @@ struct _CompIcon {
     int		height;
 };
 
-typedef struct _CompOutput {
+struct _CompOutput {
     char       *name;
+    int        id;
     REGION     region;
     int        width;
     int        height;
     XRectangle workArea;
-} CompOutput;
+};
+
+typedef struct _CompCursorImage {
+    struct _CompCursorImage *next;
+
+    unsigned long serial;
+    Pixmap	  pixmap;
+    CompTexture   texture;
+    int		  xhot;
+    int	          yhot;
+    int		  width;
+    int	          height;
+} CompCursorImage;
+
+struct _CompCursor {
+    struct _CompCursor *next;
+
+    CompScreen	    *screen;
+    CompCursorImage *image;
+
+    int	x;
+    int	y;
+
+    CompMatrix matrix;
+};
+
+#define ACTIVE_WINDOW_HISTORY_SIZE 64
+#define ACTIVE_WINDOW_HISTORY_NUM  32
+
+typedef struct _CompActiveWindowHistory {
+    Window id[ACTIVE_WINDOW_HISTORY_SIZE];
+    int    x;
+    int    y;
+    int    activeNum;
+} CompActiveWindowHistory;
 
 struct _CompScreen {
     CompScreen  *next;
@@ -1614,6 +1881,12 @@ struct _CompScreen {
     CompOutput *outputDev;
     int	       nOutputDev;
     int	       currentOutputDev;
+    CompOutput fullscreenOutput;
+
+    XRectangle lastViewport;
+
+    CompActiveWindowHistory history[ACTIVE_WINDOW_HISTORY_NUM];
+    int			    currentHistory;
 
     int overlayWindowCount;
 
@@ -1677,8 +1950,6 @@ struct _CompScreen {
     Bool lighting;
     Bool slowAnimations;
 
-    int opacityStep;
-
     XRectangle workArea;
 
     unsigned int showingDesktopMask;
@@ -1686,6 +1957,9 @@ struct _CompScreen {
     unsigned long *desktopHintData;
     int           desktopHintSize;
 
+    CompCursor      *cursors;
+    CompCursorImage *cursorImages;
+
     GLXGetProcAddressProc    getProcAddress;
     GLXBindTexImageProc      bindTexImage;
     GLXReleaseTexImageProc   releaseTexImage;
@@ -1724,29 +1998,41 @@ struct _CompScreen {
     InitPluginForScreenProc initPluginForScreen;
     FiniPluginForScreenProc finiPluginForScreen;
 
-    PreparePaintScreenProc	  preparePaintScreen;
-    DonePaintScreenProc		  donePaintScreen;
-    PaintScreenProc		  paintScreen;
-    PaintTransformedScreenProc	  paintTransformedScreen;
-    ApplyScreenTransformProc	  applyScreenTransform;
-    PaintBackgroundProc		  paintBackground;
-    PaintWindowProc		  paintWindow;
-    DrawWindowProc		  drawWindow;
-    AddWindowGeometryProc	  addWindowGeometry;
-    DrawWindowTextureProc	  drawWindowTexture;
-    DamageWindowRectProc	  damageWindowRect;
-    GetOutputExtentsForWindowProc getOutputExtentsForWindow;
-    FocusWindowProc		  focusWindow;
+    PreparePaintScreenProc	   preparePaintScreen;
+    DonePaintScreenProc		   donePaintScreen;
+    PaintScreenProc		   paintScreen;
+    PaintOutputProc		   paintOutput;
+    PaintTransformedOutputProc	   paintTransformedOutput;
+    ApplyScreenTransformProc	   applyScreenTransform;
+    PaintBackgroundProc		   paintBackground;
+    PaintWindowProc		   paintWindow;
+    DrawWindowProc		   drawWindow;
+    AddWindowGeometryProc	   addWindowGeometry;
+    DrawWindowTextureProc	   drawWindowTexture;
+    DamageWindowRectProc	   damageWindowRect;
+    GetOutputExtentsForWindowProc  getOutputExtentsForWindow;
+    GetAllowedActionsForWindowProc getAllowedActionsForWindow;
+    FocusWindowProc		   focusWindow;
+    PlaceWindowProc                placeWindow;
 
+    PaintCursorProc      paintCursor;
+    DamageCursorRectProc damageCursorRect;
+
+    WindowAddNotifyProc    windowAddNotify;
     WindowResizeNotifyProc windowResizeNotify;
     WindowMoveNotifyProc   windowMoveNotify;
     WindowGrabNotifyProc   windowGrabNotify;
     WindowUngrabNotifyProc windowUngrabNotify;
 
+    EnterShowDesktopModeProc enterShowDesktopMode;
+    LeaveShowDesktopModeProc leaveShowDesktopMode;
+
     WindowStateChangeNotifyProc windowStateChangeNotify;
 
     OutputChangeNotifyProc outputChangeNotify;
 
+    InitWindowWalkerProc initWindowWalker;
+
     CompPrivate *privates;
 };
 
@@ -1882,6 +2168,7 @@ moveScreenViewport (CompScreen *s,
 void
 moveWindowToViewportPosition (CompWindow *w,
 			      int	 x,
+			      int        y,
 			      Bool       sync);
 
 CompGroup *
@@ -1899,13 +2186,6 @@ void
 applyStartupProperties (CompScreen *screen,
 			CompWindow *window);
 
-void
-enterShowDesktopMode (CompScreen *s);
-
-void
-leaveShowDesktopMode (CompScreen *s,
-		      CompWindow *window);
-
 void
 sendWindowActivationRequest (CompScreen *s,
 			     Window	id);
@@ -1968,7 +2248,7 @@ outputChangeNotify (CompScreen *s);
 
 void
 clearScreenOutput (CompScreen   *s,
-		   int	        output,
+		   CompOutput	*output,
 		   unsigned int mask);
 
 void
@@ -1992,6 +2272,22 @@ outputDeviceForGeometry (CompScreen *s,
 Bool
 updateDefaultIcon (CompScreen *screen);
 
+CompCursor *
+findCursorAtScreen (CompScreen *screen);
+
+CompCursorImage *
+findCursorImageAtScreen (CompScreen    *screen,
+			 unsigned long serial);
+
+void
+setCurrentActiveWindowHistory (CompScreen *s,
+			       int	  x,
+			       int	  y);
+
+void
+addToCurrentActiveWindowHistory (CompScreen *s,
+				 Window	    id);
+
 
 /* window.c */
 
@@ -2008,6 +2304,12 @@ typedef Bool (*InitPluginForWindowProc) (CompPlugin *plugin,
 typedef void (*FiniPluginForWindowProc) (CompPlugin *plugin,
 					 CompWindow *window);
 
+typedef enum {
+    CompStackingUpdateModeNone = 0,
+    CompStackingUpdateModeNormal,
+    CompStackingUpdateModeAboveFullscreen
+} CompStackingUpdateMode;
+
 struct _CompWindowExtents {
     int left;
     int right;
@@ -2054,6 +2356,7 @@ struct _CompWindow {
     unsigned int      wmType;
     unsigned int      type;
     unsigned int      state;
+    unsigned int      lastState;
     unsigned int      actions;
     unsigned int      protocols;
     unsigned int      mwmDecor;
@@ -2064,6 +2367,7 @@ struct _CompWindow {
     Bool	      redirected;
     Bool	      managed;
     Bool	      bindFailed;
+    Bool	      added;
     int		      destroyRefCnt;
     int		      unmapRefCnt;
 
@@ -2078,6 +2382,7 @@ struct _CompWindow {
     Bool inShowDesktopMode;
     Bool shaded;
     Bool hidden;
+    Bool grabbed;
 
     unsigned int desktop;
 
@@ -2097,9 +2402,14 @@ struct _CompWindow {
     GLushort brightness;
     GLushort saturation;
 
+    Bool opacityPropSet;
+    int  opacityFactor;
+
     WindowPaintAttrib paint;
     WindowPaintAttrib lastPaint;
 
+    unsigned int lastMask;
+
     CompWindowExtents input;
     CompWindowExtents output;
 
@@ -2157,6 +2467,9 @@ unsigned int
 windowStateMask (CompDisplay *display,
 		 Atom	     state);
 
+unsigned int
+windowStateFromString (const char *str);
+
 unsigned int
 getWindowState (CompDisplay *display,
 		Window      id);
@@ -2166,6 +2479,10 @@ setWindowState (CompDisplay  *display,
 		unsigned int state,
 		Window       id);
 
+void
+changeWindowState (CompWindow   *w,
+		   unsigned int newState);
+
 void
 recalcWindowActions (CompWindow *w);
 
@@ -2173,6 +2490,9 @@ unsigned int
 constrainWindowState (unsigned int state,
 		      unsigned int actions);
 
+unsigned int
+windowTypeFromString (const char *str);
+
 unsigned int
 getWindowType (CompDisplay *display,
 	       Window      id);
@@ -2202,6 +2522,12 @@ setWindowProp (CompDisplay  *display,
 	       Atom	    property,
 	       unsigned int value);
 
+Bool
+readWindowProp32 (CompDisplay    *display,
+		  Window	 id,
+		  Atom		 property,
+		  unsigned short *returnValue);
+
 unsigned short
 getWindowProp32 (CompDisplay	*display,
 		 Window		id,
@@ -2214,6 +2540,9 @@ setWindowProp32 (CompDisplay    *display,
 		 Atom		property,
 		 unsigned short value);
 
+void
+updateWindowOpacity (CompWindow *window);
+
 void
 updateNormalHints (CompWindow *window);
 
@@ -2328,6 +2657,9 @@ void
 getOutputExtentsForWindow (CompWindow	     *w,
 			   CompWindowExtents *output);
 
+unsigned int
+getAllowedActionsForWindow (CompWindow *w);
+
 void
 addWindowDamage (CompWindow *w);
 
@@ -2350,8 +2682,19 @@ damageTransformedWindowRect (CompWindow *w,
 Bool
 focusWindow (CompWindow *w);
 
+Bool
+placeWindow (CompWindow *w,
+	     int        x,
+	     int        y,
+	     int        *newX,
+	     int        *newY);
+
 void
-windowResizeNotify (CompWindow *w);
+windowResizeNotify (CompWindow *w,
+		    int	       dx,
+		    int	       dy,
+		    int	       dwidth,
+		    int	       dheight);
 
 void
 windowMoveNotify (CompWindow *w,
@@ -2393,8 +2736,8 @@ restackWindowBelow (CompWindow *w,
 		    CompWindow *sibling);
 
 void
-updateWindowAttributes (CompWindow *w,
-			Bool	   aboveFs);
+updateWindowAttributes (CompWindow             *w,
+			CompStackingUpdateMode stackingMode);
 
 void
 activateWindow (CompWindow *w);
@@ -2403,10 +2746,6 @@ void
 closeWindow (CompWindow *w,
 	     Time	serverTime);
 
-void
-getOuterRectOfWindow (CompWindow *w,
-		      XRectangle *r);
-
 Bool
 constrainNewWindowSize (CompWindow *w,
 			int        width,
@@ -2470,6 +2809,13 @@ void
 setDesktopForWindow (CompWindow   *w,
 		     unsigned int desktop);
 
+int
+compareWindowActiveness (CompWindow *w1,
+			 CompWindow *w2);
+
+void
+windowAddNotify (CompWindow *w);
+
 
 /* plugin.c */
 
@@ -2481,27 +2827,13 @@ typedef int (*GetVersionProc) (CompPlugin *plugin,
 typedef Bool (*InitPluginProc) (CompPlugin *plugin);
 typedef void (*FiniPluginProc) (CompPlugin *plugin);
 
-typedef enum {
-    CompPluginRuleBefore,
-    CompPluginRuleAfter,
-    CompPluginRuleRequire
-} CompPluginRule;
-
-typedef struct _CompPluginDep {
-    CompPluginRule rule;
-    char	   *name;
-} CompPluginDep;
-
-typedef struct _CompPluginFeature {
-    char *name;
-} CompPluginFeature;
+typedef CompMetadata *(*GetMetadataProc) (CompPlugin *plugin);
 
 typedef struct _CompPluginVTable {
     char *name;
-    char *shortDesc;
-    char *longDesc;
 
-    GetVersionProc getVersion;
+    GetVersionProc  getVersion;
+    GetMetadataProc getMetadata;
 
     InitPluginProc init;
     FiniPluginProc fini;
@@ -2515,16 +2847,10 @@ typedef struct _CompPluginVTable {
     InitPluginForWindowProc initWindow;
     FiniPluginForWindowProc finiWindow;
 
-    GetDisplayOptionsProc getDisplayOptions;
-    SetDisplayOptionProc  setDisplayOption;
-    GetScreenOptionsProc  getScreenOptions;
-    SetScreenOptionProc   setScreenOption;
-
-    CompPluginDep *deps;
-    int		  nDeps;
-
-    CompPluginFeature *features;
-    int		      nFeatures;
+    GetPluginDisplayOptionsProc getDisplayOptions;
+    SetPluginDisplayOptionProc  setDisplayOption;
+    GetPluginScreenOptionsProc  getScreenOptions;
+    SetPluginScreenOptionProc   setScreenOption;
 } CompPluginVTable;
 
 typedef CompPluginVTable *(*PluginGetInfoProc) (void);
@@ -2583,9 +2909,6 @@ windowFiniPlugins (CompWindow *w);
 CompPlugin *
 findActivePlugin (char *name);
 
-CompPluginFeature *
-findActiveFeature (char *name);
-
 CompPlugin *
 loadPlugin (char *plugin);
 
@@ -2663,11 +2986,13 @@ addColorOpToFunctionData (CompFunctionData *data,
 
 Bool
 addDataOpToFunctionData (CompFunctionData *data,
-			 char		  *str);
+			 char		  *str,
+			 ...);
 
 Bool
 addBlendOpToFunctionData (CompFunctionData *data,
-			  char		   *str);
+			  char		   *str,
+			  ...);
 
 int
 createFragmentFunction (CompScreen	 *s,
@@ -2730,6 +3055,238 @@ matrixTranslate (CompTransform *transform,
 		 float	       y,
 		 float	       z);
 
+
+/* cursor.c */
+
+void
+addCursor (CompScreen *s);
+
+Bool
+damageCursorRect (CompCursor *c,
+		  Bool       initial,
+		  BoxPtr     rect);
+
+void
+addCursorDamageRect (CompCursor *c,
+		     BoxPtr     rect);
+
+void
+addCursorDamage (CompCursor *c);
+
+void
+updateCursor (CompCursor    *c,
+	      int	    x,
+	      int	    y,
+	      unsigned long serial);
+
+
+/* match.c */
+
+void
+matchInit (CompMatch *match);
+
+void
+matchFini (CompMatch *match);
+
+Bool
+matchEqual (CompMatch *m1,
+	    CompMatch *m2);
+
+Bool
+matchCopy (CompMatch *dst,
+	   CompMatch *src);
+
+Bool
+matchAddGroup (CompMatch *match,
+	       int	 flags,
+	       CompMatch *group);
+
+Bool
+matchAddExp (CompMatch *match,
+	     int	flags,
+	     const char	*value);
+
+void
+matchAddFromString (CompMatch  *match,
+		    const char *str);
+
+char *
+matchToString (CompMatch *match);
+
+void
+matchUpdate (CompDisplay *display,
+	     CompMatch   *match);
+
+Bool
+matchEval (CompMatch  *match,
+	   CompWindow *window);
+
+void
+matchInitExp (CompDisplay  *display,
+	      CompMatchExp *exp,
+	      const char   *value);
+
+void
+matchExpHandlerChanged (CompDisplay *display);
+
+void
+matchPropertyChanged (CompDisplay *display,
+		      CompWindow  *window);
+
+
+/* metadata.c */
+
+#define STRINGIFY(x) #x
+#define TOSTRING(x) STRINGIFY (x)
+#define MINTOSTRING(x) "<min>" TOSTRING (x) "</min>"
+#define MAXTOSTRING(x) "<max>" TOSTRING (x) "</max>"
+#define RESTOSTRING(min, max) MINTOSTRING (min) MAXTOSTRING (max)
+
+typedef struct _CompMetadataOptionInfo {
+    char		   *name;
+    char		   *type;
+    char		   *data;
+    CompActionCallBackProc initiate;
+    CompActionCallBackProc terminate;
+} CompMetadataOptionInfo;
+
+extern const CompMetadataOptionInfo
+coreDisplayOptionInfo[COMP_DISPLAY_OPTION_NUM];
+extern const CompMetadataOptionInfo
+coreScreenOptionInfo[COMP_SCREEN_OPTION_NUM];
+
+struct _CompMetadata {
+    char   *path;
+    xmlDoc **doc;
+    int    nDoc;
+};
+
+Bool
+compInitMetadata (CompMetadata *metadata);
+
+Bool
+compInitPluginMetadata (CompMetadata *metadata,
+			const char   *plugin);
+
+Bool
+compInitPluginMetadataFromInfo (CompMetadata		     *metadata,
+				const char		     *plugin,
+				const CompMetadataOptionInfo *displayOptionInfo,
+				int			     nDisplayOptionInfo,
+				const CompMetadataOptionInfo *screenOptionInfo,
+				int			     nScreenOptionInfo);
+
+void
+compFiniMetadata (CompMetadata *metadata);
+
+Bool
+compAddMetadataFromFile (CompMetadata *metadata,
+			 const char   *file);
+
+Bool
+compAddMetadataFromString (CompMetadata *metadata,
+			   const char	*string);
+
+Bool
+compAddMetadataFromIO (CompMetadata	     *metadata,
+		       xmlInputReadCallback  ioread,
+		       xmlInputCloseCallback ioclose,
+		       void		     *ioctx);
+
+Bool
+compInitScreenOptionFromMetadata (CompScreen   *screen,
+				  CompMetadata *metadata,
+				  CompOption   *option,
+				  const char   *name);
+
+void
+compFiniScreenOption (CompScreen *screen,
+		      CompOption *option);
+
+Bool
+compInitScreenOptionsFromMetadata (CompScreen			*screen,
+				   CompMetadata			*metadata,
+				   const CompMetadataOptionInfo *info,
+				   CompOption			*option,
+				   int				n);
+
+void
+compFiniScreenOptions (CompScreen *screen,
+		       CompOption *option,
+		       int	  n);
+
+Bool
+compSetScreenOption (CompScreen      *screen,
+		     CompOption      *option,
+		     CompOptionValue *value);
+
+Bool
+compInitDisplayOptionFromMetadata (CompDisplay  *display,
+				   CompMetadata *metadata,
+				   CompOption   *option,
+				   const char   *name);
+
+void
+compFiniDisplayOption (CompDisplay *display,
+		       CompOption  *option);
+
+Bool
+compInitDisplayOptionsFromMetadata (CompDisplay			 *display,
+				    CompMetadata		 *metadata,
+				    const CompMetadataOptionInfo *info,
+				    CompOption			 *option,
+				    int				 n);
+
+void
+compFiniDisplayOptions (CompDisplay *display,
+			CompOption  *option,
+			int	    n);
+
+Bool
+compSetDisplayOption (CompDisplay     *display,
+		      CompOption      *option,
+		      CompOptionValue *value);
+
+char *
+compGetStringFromMetadataPath (CompMetadata *metadata,
+			       const char   *path);
+
+char *
+compGetShortPluginDescription (CompMetadata *metadata);
+
+char *
+compGetLongPluginDescription (CompMetadata *metadata);
+
+char *
+compGetShortScreenOptionDescription (CompMetadata *metadata,
+				     CompOption   *option);
+
+char *
+compGetLongScreenOptionDescription (CompMetadata *metadata,
+				    CompOption   *option);
+
+char *
+compGetShortDisplayOptionDescription (CompMetadata *metadata,
+				      CompOption   *option);
+
+char *
+compGetLongDisplayOptionDescription (CompMetadata *metadata,
+				     CompOption   *option);
+
+
+int
+compReadXmlChunk (const char *src,
+		  int	     *offset,
+		  char	     *buffer,
+		  int	     length);
+
+int
+compReadXmlChunkFromMetadataOptionInfo (const CompMetadataOptionInfo *info,
+					int			     *offset,
+					char			     *buffer,
+					int			     length);
+
+
 #ifdef  __cplusplus
 }
 #endif
