@@ -32,6 +32,7 @@
 #include <klocale.h>
 #include <kcommondecoration.h>
 #include <kwin.h>
+#include <qwidgetlist.h>
 #include <qpoint.h>
 
 #include <X11/Xlib.h>
@@ -60,6 +61,7 @@ double decorationOpacity = 0.75;
 bool   decorationOpacityShade = false;
 double activeDecorationOpacity = 1.0;
 bool   activeDecorationOpacityShade = false;
+int    blurType = BLUR_TYPE_NONE;
 
 decor_context_t KWD::Decorator::mDefaultContext;
 decor_shadow_t *KWD::Decorator::mNoBorderShadow = 0;
@@ -190,8 +192,7 @@ KWD::Decorator::Decorator (void) : DCOPObject ("KWinInterface"),
     mConfig (0),
     mKWinModule (new KWinModule (this, KWinModule::INFO_ALL)),
     mDBusQtConnection (this),
-    mCompositeWindow (0, 0, Qt::WType_TopLevel | Qt::WStyle_NoBorder |
-		      Qt::WX11BypassWM)
+    mCompositeWindow (0)
 {
     DCOPClient *client;
     int	       i, j;
@@ -229,12 +230,15 @@ KWD::Decorator::Decorator (void) : DCOPObject ("KWinInterface"),
     mShadowOptions.shadow_color[0] = SHADOW_COLOR_RED;
     mShadowOptions.shadow_color[1] = SHADOW_COLOR_GREEN;
     mShadowOptions.shadow_color[2] = SHADOW_COLOR_BLUE;
-    mShadowOptions.shadow_color[3] = 0xffff;
 
-    mCompositeWindow.setGeometry (QRect (-ROOT_OFF_X, -ROOT_OFF_Y, 1, 1));
-    mCompositeWindow.show ();
+    mCompositeWindow = new QWidget (0, "KWDCompositeWidget",
+				    Qt::WType_TopLevel | Qt::WStyle_NoBorder |
+				    Qt::WX11BypassWM);
 
-    XCompositeRedirectSubwindows (qt_xdisplay (), mCompositeWindow.winId (),
+    mCompositeWindow->setGeometry (QRect (-ROOT_OFF_X, -ROOT_OFF_Y, 1, 1));
+    mCompositeWindow->show ();
+
+    XCompositeRedirectSubwindows (qt_xdisplay (), mCompositeWindow->winId (),
 				  CompositeRedirectManual);
 }
 
@@ -251,6 +255,9 @@ KWD::Decorator::~Decorator (void)
     if (mDecorActive)
 	delete mDecorActive;
 
+    /* XXX: mCompositeWindow is not deleted, some plugins seem to rely on
+       it not being deleted... not sure what to do about this. */
+
     delete mOptions;
     delete mPlugins;
     delete mConfig;
@@ -291,9 +298,9 @@ KWD::Decorator::enableDecorations (Time timestamp,
 
     updateShadow ();
 
-    mDecorNormal = new KWD::Window (&mCompositeWindow, qt_xrootwin (),
+    mDecorNormal = new KWD::Window (mCompositeWindow, qt_xrootwin (),
 				    0, Window::Default);
-    mDecorActive = new KWD::Window (&mCompositeWindow, qt_xrootwin (),
+    mDecorActive = new KWD::Window (mCompositeWindow, qt_xrootwin (),
 				    0, Window::DefaultActive);
 
     connect (mKWinModule, SIGNAL (windowAdded (WId)),
@@ -350,7 +357,7 @@ KWD::Decorator::updateDefaultShadow (KWD::Window *w)
 }
 
 DBusMessage *
-KWD::Decorator::sendAndBlockForShadowOptionReply (char *path)
+KWD::Decorator::sendAndBlockForShadowOptionReply (const char *path)
 {
     DBusMessage *message;
 
@@ -728,7 +735,7 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
     case ClientMessage:
 	if (xevent->xclient.message_type == Atoms::toolkitActionAtom)
 	{
-	    long action;
+	    unsigned long action;
 
 	    action = xevent->xclient.data.l[0];
 	    if (action == Atoms::toolkitActionWindowMenuAtom)
@@ -850,9 +857,25 @@ KWD::Decorator::handleWindowAdded (WId id)
     unsigned int			     width, height, border, depth;
     int					     x, y;
     XID					     root;
+    QWidgetList				     *widgets;
+
+    /* avoid adding any of our own top level windows */
+    widgets = QApplication::topLevelWidgets ();
+    if (widgets)
+    {
+	for (QWidgetListIt it (*widgets); it.current (); ++it)
+	{
+	    if (it.current ()->winId () == id)
+	    {
+		delete widgets;
+		return;
+	    }
+	}
+
+	delete widgets;
+    }
 
     KWD::trapXError ();
-    XSelectInput (qt_xdisplay (), id, StructureNotifyMask | PropertyChangeMask);
     XGetGeometry (qt_xdisplay (), id, &root, &x, &y, &width, &height,
 		  &border, &depth);
     if (KWD::popXError ())
@@ -885,11 +908,15 @@ KWD::Decorator::handleWindowAdded (WId id)
 	type = KWD::Window::Normal;
     }
 
+    KWD::trapXError ();
+    XSelectInput (qt_xdisplay (), id, StructureNotifyMask | PropertyChangeMask);
+    KWD::popXError ();
+
     if (frame)
     {
 	if (!mClients.contains (id))
 	{
-	    client = new KWD::Window (&mCompositeWindow, id, frame, type,
+	    client = new KWD::Window (mCompositeWindow, id, frame, type,
 				      x, y,
 				      width + border * 2,
 				      height + border * 2);
@@ -911,7 +938,7 @@ KWD::Decorator::handleWindowAdded (WId id)
     {
 	if (!mClients.contains (id))
 	{
-	    client = new KWD::Window (&mCompositeWindow, id, 0, type,
+	    client = new KWD::Window (mCompositeWindow, id, 0, type,
 				      x, y,
 				      width + border * 2,
 				      height + border * 2);
