@@ -37,30 +37,9 @@
 #include <X11/extensions/shape.h>
 
 #include <compiz.h>
+#include <decoration.h>
 
-#define GRAVITY_WEST  (1 << 0)
-#define GRAVITY_EAST  (1 << 1)
-#define GRAVITY_NORTH (1 << 2)
-#define GRAVITY_SOUTH (1 << 3)
-
-#define ALIGN_LEFT   (0)
-#define ALIGN_RIGHT  (1 << 0)
-#define ALIGN_TOP    (0)
-#define ALIGN_BOTTOM (1 << 1)
-
-#define CLAMP_HORZ (1 << 0)
-#define CLAMP_VERT (1 << 1)
-
-#define XX_MASK (1 << 12)
-#define XY_MASK (1 << 13)
-#define YX_MASK (1 << 14)
-#define YY_MASK (1 << 15)
-
-typedef struct _Point {
-    int	x;
-    int	y;
-    int gravity;
-} Point;
+static CompMetadata decorMetadata;
 
 typedef struct _Vector {
     int	dx;
@@ -69,16 +48,6 @@ typedef struct _Vector {
     int	y0;
 } Vector;
 
-typedef struct _Quad {
-    Point      p1;
-    Point      p2;
-    int	       maxWidth;
-    int	       maxHeight;
-    int	       align;
-    int	       clamp;
-    CompMatrix m;
-} Quad;
-
 #define DECOR_BARE   0
 #define DECOR_NORMAL 1
 #define DECOR_ACTIVE 2
@@ -100,13 +69,15 @@ typedef struct _Decoration {
     CompWindowExtents maxInput;
     int		      minWidth;
     int		      minHeight;
-    Quad	      *quad;
+    decor_quad_t      *quad;
     int		      nQuad;
 } Decoration;
 
 typedef struct _ScaledQuad {
     CompMatrix matrix;
     BoxRec     box;
+    float      sx;
+    float      sy;
 } ScaledQuad;
 
 typedef struct _WindowDecoration {
@@ -115,41 +86,27 @@ typedef struct _WindowDecoration {
     int	       nQuad;
 } WindowDecoration;
 
-#define DECOR_SHADOW_RADIUS_DEFAULT    8.0f
-#define DECOR_SHADOW_RADIUS_MIN        0.0f
-#define DECOR_SHADOW_RADIUS_MAX       48.0f
-#define DECOR_SHADOW_RADIUS_PRECISION  0.1f
-
-#define DECOR_SHADOW_OPACITY_DEFAULT   0.5f
-#define DECOR_SHADOW_OPACITY_MIN       0.01f
-#define DECOR_SHADOW_OPACITY_MAX       6.0f
-#define DECOR_SHADOW_OPACITY_PRECISION 0.01f
-
-#define DECOR_SHADOW_COLOR_RED_DEFAULT   0x0000
-#define DECOR_SHADOW_COLOR_GREEN_DEFAULT 0x0000
-#define DECOR_SHADOW_COLOR_BLUE_DEFAULT  0x0000
-
-#define DECOR_SHADOW_OFFSET_DEFAULT   1
-#define DECOR_SHADOW_OFFSET_MIN     -16
-#define DECOR_SHADOW_OFFSET_MAX      16
-
 #define DECOR_DISPLAY_OPTION_SHADOW_RADIUS   0
 #define DECOR_DISPLAY_OPTION_SHADOW_OPACITY  1
 #define DECOR_DISPLAY_OPTION_SHADOW_COLOR    2
 #define DECOR_DISPLAY_OPTION_SHADOW_OFFSET_X 3
 #define DECOR_DISPLAY_OPTION_SHADOW_OFFSET_Y 4
 #define DECOR_DISPLAY_OPTION_COMMAND         5
-#define DECOR_DISPLAY_OPTION_NUM             6
+#define DECOR_DISPLAY_OPTION_MIPMAP          6
+#define DECOR_DISPLAY_OPTION_DECOR_MATCH     7
+#define DECOR_DISPLAY_OPTION_SHADOW_MATCH    8
+#define DECOR_DISPLAY_OPTION_NUM             9
 
 static int displayPrivateIndex;
 
 typedef struct _DecorDisplay {
-    int		    screenPrivateIndex;
-    HandleEventProc handleEvent;
-    DecorTexture    *textures;
-    Atom	    supportingDmCheckAtom;
-    Atom	    winDecorAtom;
-    Atom	    decorAtom[DECOR_NUM];
+    int			     screenPrivateIndex;
+    HandleEventProc	     handleEvent;
+    MatchPropertyChangedProc matchPropertyChanged;
+    DecorTexture	     *textures;
+    Atom		     supportingDmCheckAtom;
+    Atom		     winDecorAtom;
+    Atom		     decorAtom[DECOR_NUM];
 
     CompOption opt[DECOR_DISPLAY_OPTION_NUM];
 } DecorDisplay;
@@ -161,9 +118,11 @@ typedef struct _DecorScreen {
 
     Decoration *decor[DECOR_NUM];
 
-    PaintWindowProc	 paintWindow;
-    DamageWindowRectProc damageWindowRect;
+    DrawWindowProc		  drawWindow;
+    DamageWindowRectProc	  damageWindowRect;
+    GetOutputExtentsForWindowProc getOutputExtentsForWindow;
 
+    WindowAddNotifyProc    windowAddNotify;
     WindowMoveNotifyProc   windowMoveNotify;
     WindowResizeNotifyProc windowResizeNotify;
 
@@ -197,191 +156,57 @@ typedef struct _DecorWindow {
 
 #define NUM_OPTIONS(d) (sizeof ((d)->opt) / sizeof (CompOption))
 
-static CompOption *
-decorGetDisplayOptions (CompDisplay *display,
-			int	    *count)
-{
-    DECOR_DISPLAY (display);
-
-    *count = NUM_OPTIONS (dd);
-    return dd->opt;
-}
-
 static Bool
-decorSetDisplayOption (CompDisplay     *display,
-		       char	       *name,
-		       CompOptionValue *value)
-{
-    CompOption *o;
-    int	       index;
-
-    DECOR_DISPLAY (display);
-
-    o = compFindOption (dd->opt, NUM_OPTIONS (dd), name, &index);
-    if (!o)
-	return FALSE;
-
-    switch (index) {
-    case DECOR_DISPLAY_OPTION_SHADOW_RADIUS:
-    case DECOR_DISPLAY_OPTION_SHADOW_OPACITY:
-	if (compSetFloatOption (o, value))
-	    return TRUE;
-	break;
-    case DECOR_DISPLAY_OPTION_SHADOW_COLOR:
-	if (compSetColorOption (o, value))
-	    return TRUE;
-	break;
-    case DECOR_DISPLAY_OPTION_SHADOW_OFFSET_X:
-    case DECOR_DISPLAY_OPTION_SHADOW_OFFSET_Y:
-	if (compSetIntOption (o, value))
-	    return TRUE;
-	break;
-    case DECOR_DISPLAY_OPTION_COMMAND:
-	if (compSetStringOption (o, value))
-	{
-	    if (display->screens && *o->value.s != '\0')
-	    {
-		DECOR_SCREEN (display->screens);
-
-		/* run decorator command if no decorator is present on
-		   first screen */
-		if (!ds->dmWin)
-		{
-		    if (fork () == 0)
-		    {
-			putenv (display->displayString);
-			execl ("/bin/sh", "/bin/sh", "-c", o->value.s, NULL);
-			exit (0);
-		    }
-		}
-	    }
-
-	    return TRUE;
-	}
-    default:
-	break;
-    }
-
-    return FALSE;
-}
-
-static void
-decorDisplayInitOptions (DecorDisplay *dd)
-{
-    CompOption *o;
-
-    o = &dd->opt[DECOR_DISPLAY_OPTION_SHADOW_RADIUS];
-    o->name		= "shadow_radius";
-    o->shortDesc	= N_("Shadow Radius");
-    o->longDesc		= N_("Drop shadow radius");
-    o->type		= CompOptionTypeFloat;
-    o->value.f		= DECOR_SHADOW_RADIUS_DEFAULT;
-    o->rest.f.min	= DECOR_SHADOW_RADIUS_MIN;
-    o->rest.f.max	= DECOR_SHADOW_RADIUS_MAX;
-    o->rest.f.precision = DECOR_SHADOW_RADIUS_PRECISION;
-
-    o = &dd->opt[DECOR_DISPLAY_OPTION_SHADOW_OPACITY];
-    o->name		= "shadow_opacity";
-    o->shortDesc	= N_("Shadow Opacity");
-    o->longDesc		= N_("Drop shadow opacity");
-    o->type		= CompOptionTypeFloat;
-    o->value.f		= DECOR_SHADOW_OPACITY_DEFAULT;
-    o->rest.f.min	= DECOR_SHADOW_OPACITY_MIN;
-    o->rest.f.max	= DECOR_SHADOW_OPACITY_MAX;
-    o->rest.f.precision = DECOR_SHADOW_OPACITY_PRECISION;
-
-    o = &dd->opt[DECOR_DISPLAY_OPTION_SHADOW_COLOR];
-    o->name		= "shadow_color";
-    o->shortDesc	= "Shadow Color";
-    o->longDesc		= "Drop shadow color";
-    o->type		= CompOptionTypeColor;
-    o->value.c[0]	= DECOR_SHADOW_COLOR_RED_DEFAULT;
-    o->value.c[1]	= DECOR_SHADOW_COLOR_GREEN_DEFAULT;
-    o->value.c[2]	= DECOR_SHADOW_COLOR_BLUE_DEFAULT;
-    o->value.c[3]	= 0xffff;
-
-    o = &dd->opt[DECOR_DISPLAY_OPTION_SHADOW_OFFSET_X];
-    o->name		= "shadow_offset_x";
-    o->shortDesc	= N_("Shadow Offset X");
-    o->longDesc		= N_("Drop shadow X offset");
-    o->type		= CompOptionTypeInt;
-    o->value.i		= DECOR_SHADOW_OFFSET_DEFAULT;
-    o->rest.i.min	= DECOR_SHADOW_OFFSET_MIN;
-    o->rest.i.max	= DECOR_SHADOW_OFFSET_MAX;
-
-    o = &dd->opt[DECOR_DISPLAY_OPTION_SHADOW_OFFSET_Y];
-    o->name		= "shadow_offset_y";
-    o->shortDesc	= N_("Shadow Offset Y");
-    o->longDesc		= N_("Drop shadow Y offset");
-    o->type		= CompOptionTypeInt;
-    o->value.i		= DECOR_SHADOW_OFFSET_DEFAULT;
-    o->rest.i.min	= DECOR_SHADOW_OFFSET_MIN;
-    o->rest.i.max	= DECOR_SHADOW_OFFSET_MAX;
-
-    o = &dd->opt[DECOR_DISPLAY_OPTION_COMMAND];
-    o->name		= "command";
-    o->shortDesc	= N_("Command");
-    o->longDesc		= N_("Decorator command line that is executed if no "
-			     "decorator is already running");
-    o->type		= CompOptionTypeString;
-    o->value.s		= strdup ("");
-    o->rest.s.string	= NULL;
-    o->rest.s.nString	= 0;
-}
-
-static Bool
-decorPaintWindow (CompWindow		  *w,
-		  const WindowPaintAttrib *attrib,
-		  Region		  region,
-		  unsigned int		  mask)
+decorDrawWindow (CompWindow	      *w,
+		 const CompTransform  *transform,
+		 const FragmentAttrib *attrib,
+		 Region		      region,
+		 unsigned int	      mask)
 {
     Bool status;
 
     DECOR_SCREEN (w->screen);
+    DECOR_WINDOW (w);
 
-    if (!(mask & PAINT_WINDOW_SOLID_MASK))
+    UNWRAP (ds, w->screen, drawWindow);
+    status = (*w->screen->drawWindow) (w, transform, attrib, region, mask);
+    WRAP (ds, w->screen, drawWindow, decorDrawWindow);
+
+    if (mask & PAINT_WINDOW_TRANSFORMED_MASK)
+	region = &infiniteRegion;
+
+    if (dw->wd && region->numRects)
     {
-	DECOR_WINDOW (w);
+	WindowDecoration *wd = dw->wd;
+	REGION	     box;
+	int		     i;
 
-	if (mask & PAINT_WINDOW_TRANSFORMED_MASK)
-	    region = &infiniteRegion;
+	mask |= PAINT_WINDOW_BLEND_MASK;
 
-	if (dw->wd && region->numRects)
+	box.rects	 = &box.extents;
+	box.numRects = 1;
+
+	w->vCount = w->indexCount = 0;
+
+	for (i = 0; i < wd->nQuad; i++)
 	{
-	    WindowDecoration *wd = dw->wd;
-	    REGION	     box;
-	    int		     i;
+	    box.extents = wd->quad[i].box;
 
-	    box.rects	 = &box.extents;
-	    box.numRects = 1;
-
-	    w->vCount = 0;
-
-	    for (i = 0; i < wd->nQuad; i++)
+	    if (box.extents.x1 < box.extents.x2 &&
+		box.extents.y1 < box.extents.y2)
 	    {
-		box.extents = wd->quad[i].box;
-
-		if (box.extents.x1 < box.extents.x2 &&
-		    box.extents.y1 < box.extents.y2)
-		{
-		    (*w->screen->addWindowGeometry) (w,
-						     &wd->quad[i].matrix, 1,
-						     &box,
-						     region);
-		}
+		(*w->screen->addWindowGeometry) (w,
+						 &wd->quad[i].matrix, 1,
+						 &box,
+						 region);
 	    }
-
-	    if (w->vCount)
-		(*w->screen->drawWindowTexture) (w,
-						 &wd->decor->texture->texture,
-						 attrib, mask |
-						 PAINT_WINDOW_TRANSLUCENT_MASK);
 	}
-    }
 
-    UNWRAP (ds, w->screen, paintWindow);
-    status = (*w->screen->paintWindow) (w, attrib, region, mask);
-    WRAP (ds, w->screen, paintWindow, decorPaintWindow);
+	if (w->vCount)
+	    (*w->screen->drawWindowTexture) (w,
+					     &wd->decor->texture->texture,
+					     attrib, mask);
+    }
 
     return status;
 }
@@ -428,6 +253,9 @@ decorGetTexture (CompScreen *screen,
 	return NULL;
     }
 
+    if (!dd->opt[DECOR_DISPLAY_OPTION_MIPMAP].value.b)
+	texture->texture.mipmap = FALSE;
+
     texture->damage = XDamageCreate (screen->display->display, pixmap,
 				     XDamageReportRawRectangles);
 
@@ -472,59 +300,97 @@ decorReleaseTexture (CompScreen   *screen,
     free (texture);
 }
 
-/*
-  decoration property
-  -------------------
+static void
+computeQuadBox (decor_quad_t *q,
+		int	     width,
+		int	     height,
+		int	     *return_x1,
+		int	     *return_y1,
+		int	     *return_x2,
+		int	     *return_y2,
+		float        *return_sx,
+		float        *return_sy)
+{
+    int   x1, y1, x2, y2;
+    float sx = 1.0f;
+    float sy = 1.0f;
 
-  data[0] = version
+    decor_apply_gravity (q->p1.gravity, q->p1.x, q->p1.y, width, height,
+			 &x1, &y1);
+    decor_apply_gravity (q->p2.gravity, q->p2.x, q->p2.y, width, height,
+			 &x2, &y2);
 
-  data[1] = pixmap
+    if (q->clamp & CLAMP_HORZ)
+    {
+	if (x1 < 0)
+	    x1 = 0;
+	if (x2 > width)
+	    x2 = width;
+    }
 
-  data[2] = input left
-  data[3] = input right
-  data[4] = input top
-  data[5] = input bottom
+    if (q->clamp & CLAMP_VERT)
+    {
+	if (y1 < 0)
+	    y1 = 0;
+	if (y2 > height)
+	    y2 = height;
+    }
 
-  data[6] = input left when maximized
-  data[7] = input right when maximized
-  data[8] = input top when maximized
-  data[9] = input bottom when maximized
+    if (q->stretch & STRETCH_X)
+    {
+	sx = (float)q->max_width / ((float)(x2 - x1));
+    }
+    else if (q->max_width < x2 - x1)
+    {
+	if (q->align & ALIGN_RIGHT)
+	    x1 = x2 - q->max_width;
+	else
+	    x2 = x1 + q->max_width;
+    }
 
-  data[10] = min width
-  data[11] = min height
+    if (q->stretch & STRETCH_Y)
+    {
+	sy = (float)q->max_height / ((float)(y2 - y1));
+    }
+    else if (q->max_height < y2 - y1)
+    {
+	if (q->align & ALIGN_BOTTOM)
+	    y1 = y2 - q->max_height;
+	else
+	    y2 = y1 + q->max_height;
+    }
 
-  flags
+    *return_x1 = x1;
+    *return_y1 = y1;
+    *return_x2 = x2;
+    *return_y2 = y2;
 
-  1st to 4nd bit p1 gravity, 5rd to 8th bit p2 gravity,
-  9rd and 10th bit alignment, 11rd and 12th bit clamp,
-  13th bit XX, 14th bit XY, 15th bit YX, 16th bit YY.
+    if (return_sx)
+	*return_sx = sx;
+    if (return_sy)
+	*return_sy = sy;
+}
 
-  data[11 + n * 9 + 1] = flags
-  data[11 + n * 9 + 2] = p1 x
-  data[11 + n * 9 + 3] = p1 y
-  data[11 + n * 9 + 4] = p2 x
-  data[11 + n * 9 + 5] = p2 y
-  data[11 + n * 9 + 6] = widthMax
-  data[11 + n * 9 + 7] = heightMax
-  data[11 + n * 9 + 8] = x0
-  data[11 + n * 9 + 9] = y0
- */
 static Decoration *
 decorCreateDecoration (CompScreen *screen,
 		       Window	  id,
 		       Atom	  decorAtom)
 {
-    Decoration	  *decoration;
-    Atom	  actual;
-    int		  result, format;
-    unsigned long n, nleft;
-    unsigned char *data;
-    long	  *prop;
-    Pixmap	  pixmap;
-    Quad	  *quad;
-    int		  nQuad;
-    int		  left, right, top, bottom;
-    int		  flags;
+    Decoration	    *decoration;
+    Atom	    actual;
+    int		    result, format;
+    unsigned long   n, nleft;
+    unsigned char   *data;
+    long	    *prop;
+    Pixmap	    pixmap;
+    decor_extents_t input;
+    decor_extents_t maxInput;
+    decor_quad_t    *quad;
+    int		    nQuad;
+    int		    minWidth;
+    int		    minHeight;
+    int		    left, right, top, bottom;
+    int		    x1, y1, x2, y2;
 
     result = XGetWindowProperty (screen->display->display, id,
 				 decorAtom, 0L, 1024L, FALSE,
@@ -534,116 +400,101 @@ decorCreateDecoration (CompScreen *screen,
     if (result != Success || !n || !data)
 	return NULL;
 
-    if (n < 12 + 9)
-    {
-	XFree (data);
-	return NULL;
-    }
-
     prop = (long *) data;
 
-    if (*prop != DECOR_INTERFACE_VERSION)
+    if (decor_property_get_version (prop) != decor_version ())
     {
-	fprintf (stderr, "%s: decoration: property ignored because "
-		 "version is %d and decoration plugin version is %d\n",
-		 programName, (int) *prop, DECOR_INTERFACE_VERSION);
+	compLogMessage (screen->display, "decoration", CompLogLevelWarn,
+			"Property ignored because "
+			"version is %d and decoration plugin version is %d\n",
+			decor_property_get_version (prop), decor_version ());
 
 	XFree (data);
 	return NULL;
     }
 
-    prop++;
+    nQuad = (n - BASE_PROP_SIZE) / QUAD_PROP_SIZE;
+
+    quad = malloc (sizeof (decor_quad_t) * nQuad);
+    if (!quad)
+    {
+	XFree (data);
+	return NULL;
+    }
+
+    nQuad = decor_property_to_quads (prop,
+				     n,
+				     &pixmap,
+				     &input,
+				     &maxInput,
+				     &minWidth,
+				     &minHeight,
+				     quad);
+
+    XFree (data);
+
+    if (!nQuad)
+    {
+	free (quad);
+	return NULL;
+    }
 
     decoration = malloc (sizeof (Decoration));
     if (!decoration)
     {
-	XFree (data);
+	free (quad);
 	return NULL;
     }
 
-    memcpy (&pixmap, prop++, sizeof (Pixmap));
-
     decoration->texture = decorGetTexture (screen, pixmap);
     if (!decoration->texture)
     {
 	free (decoration);
-	XFree (data);
+	free (quad);
 	return NULL;
     }
 
-    decoration->input.left   = *prop++;
-    decoration->input.right  = *prop++;
-    decoration->input.top    = *prop++;
-    decoration->input.bottom = *prop++;
+    decoration->minWidth  = minWidth;
+    decoration->minHeight = minHeight;
+    decoration->quad	  = quad;
+    decoration->nQuad	  = nQuad;
 
-    decoration->maxInput.left   = *prop++;
-    decoration->maxInput.right  = *prop++;
-    decoration->maxInput.top    = *prop++;
-    decoration->maxInput.bottom = *prop++;
-
-    decoration->minWidth  = *prop++;
-    decoration->minHeight = *prop++;
-
-    nQuad = (n - 12) / 9;
-
-    quad = malloc (sizeof (Quad) * nQuad);
-    if (!quad)
-    {
-	decorReleaseTexture (screen, decoration->texture);
-	free (decoration);
-	XFree (data);
-	return NULL;
-    }
-
-    decoration->quad  = quad;
-    decoration->nQuad = nQuad;
-
-    left = right = top = bottom = 0;
+    left   = 0;
+    right  = minWidth;
+    top    = 0;
+    bottom = minHeight;
 
     while (nQuad--)
     {
-	flags = *prop++;
+	computeQuadBox (quad, minWidth, minHeight, &x1, &y1, &x2, &y2,
+			NULL, NULL);
 
-	quad->p1.gravity = (flags >> 0) & 0xf;
-	quad->p2.gravity = (flags >> 4) & 0xf;
-
-	quad->align = (flags >> 8)  & 0x3;
-	quad->clamp = (flags >> 10) & 0x3;
-
-	quad->m.xx = (flags & XX_MASK) ? 1.0f : 0.0f;
-	quad->m.xy = (flags & XY_MASK) ? 1.0f : 0.0f;
-	quad->m.yx = (flags & YX_MASK) ? 1.0f : 0.0f;
-	quad->m.yy = (flags & YY_MASK) ? 1.0f : 0.0f;
-
-	quad->p1.x = *prop++;
-	quad->p1.y = *prop++;
-	quad->p2.x = *prop++;
-	quad->p2.y = *prop++;
-
-	quad->maxWidth  = *prop++;
-	quad->maxHeight = *prop++;
-
-	quad->m.x0 = *prop++;
-	quad->m.y0 = *prop++;
-
-	if (quad->p1.x < left)
-	    left = quad->p1.x;
-	if (quad->p1.y < top)
-	    top = quad->p1.y;
-	if (quad->p2.x > right)
-	    right = quad->p2.x;
-	if (quad->p2.y > bottom)
-	    bottom = quad->p2.y;
+	if (x1 < left)
+	    left = x1;
+	if (y1 < top)
+	    top = y1;
+	if (x2 > right)
+	    right = x2;
+	if (y2 > bottom)
+	    bottom = y2;
 
 	quad++;
     }
 
-    XFree (data);
-
     decoration->output.left   = -left;
-    decoration->output.right  = right;
+    decoration->output.right  = right - minWidth;
     decoration->output.top    = -top;
-    decoration->output.bottom = bottom;
+    decoration->output.bottom = bottom - minHeight;
+
+    decoration->input.left   = input.left;
+    decoration->input.right  = input.right;
+    decoration->input.top    = input.top;
+    decoration->input.bottom = input.bottom;
+
+    decoration->maxInput.left   = maxInput.left;
+    decoration->maxInput.right  = maxInput.right;
+    decoration->maxInput.top    = maxInput.top;
+    decoration->maxInput.bottom = maxInput.bottom;
 
     decoration->refCount = 1;
 
@@ -713,6 +564,9 @@ setDecorationMatrices (CompWindow *w)
     WindowDecoration *wd;
     int		     i;
     float	     x0, y0;
+    decor_matrix_t   a;
+    CompMatrix       b;
+
 
     DECOR_WINDOW (w);
 
@@ -725,21 +579,40 @@ setDecorationMatrices (CompWindow *w)
 	wd->quad[i].matrix = wd->decor->texture->texture.matrix;
 
 	x0 = wd->decor->quad[i].m.x0;
-	if (wd->decor->quad[i].align & ALIGN_RIGHT)
-	    x0 -= wd->quad[i].box.x2 - wd->quad[i].box.x1;
-
 	y0 = wd->decor->quad[i].m.y0;
+
+	a = wd->decor->quad[i].m;
+	b = wd->quad[i].matrix;
+
+	wd->quad[i].matrix.xx = a.xx * b.xx + a.yx * b.xy;
+	wd->quad[i].matrix.yx = a.xx * b.yx + a.yx * b.yy;
+	wd->quad[i].matrix.xy = a.xy * b.xx + a.yy * b.xy;
+	wd->quad[i].matrix.yy = a.xy * b.yx + a.yy * b.yy;
+	wd->quad[i].matrix.x0 = x0 * b.xx + y0 * b.xy + b.x0;
+	wd->quad[i].matrix.y0 = x0 * b.yx + y0 * b.yy + b.y0;
+
+	wd->quad[i].matrix.xx *= wd->quad[i].sx;
+	wd->quad[i].matrix.yx *= wd->quad[i].sx;
+	wd->quad[i].matrix.xy *= wd->quad[i].sy;
+	wd->quad[i].matrix.yy *= wd->quad[i].sy;
+
+	if (wd->decor->quad[i].align & ALIGN_RIGHT)
+	    x0 = wd->quad[i].box.x2 - wd->quad[i].box.x1;
+	else
+	    x0 = 0.0f;
+
 	if (wd->decor->quad[i].align & ALIGN_BOTTOM)
-	    y0 -= wd->quad[i].box.y2 - wd->quad[i].box.y1;
+	    y0 = wd->quad[i].box.y2 - wd->quad[i].box.y1;
+	else
+	    y0 = 0.0f;
 
-	wd->quad[i].matrix.x0 += x0 * wd->quad[i].matrix.xx;
-	wd->quad[i].matrix.y0 += y0 * wd->quad[i].matrix.yy;
+	wd->quad[i].matrix.x0 -=
+	    x0 * wd->quad[i].matrix.xx +
+	    y0 * wd->quad[i].matrix.xy;
 
-	wd->quad[i].matrix.xy = wd->decor->quad[i].m.xy * wd->quad[i].matrix.xx;
-	wd->quad[i].matrix.yx = wd->decor->quad[i].m.yx * wd->quad[i].matrix.yy;
-
-	wd->quad[i].matrix.xx *= wd->decor->quad[i].m.xx;
-	wd->quad[i].matrix.yy *= wd->decor->quad[i].m.yy;
+	wd->quad[i].matrix.y0 -=
+	    y0 * wd->quad[i].matrix.yy +
+	    x0 * wd->quad[i].matrix.yx;
 
 	wd->quad[i].matrix.x0 -=
 	    wd->quad[i].box.x1 * wd->quad[i].matrix.xx +
@@ -751,38 +624,12 @@ setDecorationMatrices (CompWindow *w)
     }
 }
 
-static void
-applyGravity (int gravity,
-	      int x,
-	      int y,
-	      int width,
-	      int height,
-	      int *return_x,
-	      int *return_y)
-{
-    if (gravity & GRAVITY_EAST)
-	*return_x = x + width;
-    else if (gravity & GRAVITY_WEST)
-	*return_x = x;
-    else
-	*return_x = (width >> 1) + x;
-
-    if (gravity & GRAVITY_SOUTH)
-	*return_y = y + height;
-    else if (gravity & GRAVITY_NORTH)
-	*return_y = y;
-    else
-	*return_y = (height >> 1) + y;
-}
-
 static void
 updateWindowDecorationScale (CompWindow *w)
 {
     WindowDecoration *wd;
     int		     x1, y1, x2, y2;
-    int		     maxWidth, maxHeight;
-    int		     align;
-    int		     clamp;
+    float            sx, sy;
     int		     i;
 
     DECOR_WINDOW (w);
@@ -793,57 +640,15 @@ updateWindowDecorationScale (CompWindow *w)
 
     for (i = 0; i < wd->nQuad; i++)
     {
-	applyGravity (wd->decor->quad[i].p1.gravity,
-		      wd->decor->quad[i].p1.x, wd->decor->quad[i].p1.y,
-		      w->width, w->height,
-		      &x1, &y1);
-
-	applyGravity (wd->decor->quad[i].p2.gravity,
-		      wd->decor->quad[i].p2.x, wd->decor->quad[i].p2.y,
-		      w->width, w->height,
-		      &x2, &y2);
-
-	maxWidth  = wd->decor->quad[i].maxWidth;
-	maxHeight = wd->decor->quad[i].maxHeight;
-	align	  = wd->decor->quad[i].align;
-	clamp	  = wd->decor->quad[i].clamp;
-
-	if (clamp & CLAMP_HORZ)
-	{
-	    if (x1 < 0)
-		x1 = 0;
-	    if (x2 > w->width)
-		x2 = w->width;
-	}
-
-	if (clamp & CLAMP_VERT)
-	{
-	    if (y1 < 0)
-		y1 = 0;
-	    if (y2 > w->height)
-		y2 = w->height;
-	}
-
-	if (maxWidth < x2 - x1)
-	{
-	    if (align & ALIGN_RIGHT)
-		x1 = x2 - maxWidth;
-	    else
-		x2 = x1 + maxWidth;
-	}
-
-	if (maxHeight < y2 - y1)
-	{
-	    if (align & ALIGN_BOTTOM)
-		y1 = y2 - maxHeight;
-	    else
-		y2 = y1 + maxHeight;
-	}
+	computeQuadBox (&wd->decor->quad[i], w->width, w->height,
+			&x1, &y1, &x2, &y2, &sx, &sy);
 
 	wd->quad[i].box.x1 = x1 + w->attrib.x;
 	wd->quad[i].box.y1 = y1 + w->attrib.y;
 	wd->quad[i].box.x2 = x2 + w->attrib.x;
 	wd->quad[i].box.y2 = y2 + w->attrib.y;
+	wd->quad[i].sx     = sx;
+	wd->quad[i].sy     = sy;
     }
 
     setDecorationMatrices (w);
@@ -856,76 +661,127 @@ decorCheckSize (CompWindow *w,
     return (decor->minWidth <= w->width && decor->minHeight <= w->height);
 }
 
+static int
+decorWindowShiftX (CompWindow *w)
+{
+    switch (w->sizeHints.win_gravity) {
+    case WestGravity:
+    case NorthWestGravity:
+    case SouthWestGravity:
+	return w->input.left;
+    case EastGravity:
+    case NorthEastGravity:
+    case SouthEastGravity:
+	return -w->input.right;
+    }
+
+    return 0;
+}
+
+static int
+decorWindowShiftY (CompWindow *w)
+{
+    switch (w->sizeHints.win_gravity) {
+    case NorthGravity:
+    case NorthWestGravity:
+    case NorthEastGravity:
+	return w->input.top;
+    case SouthGravity:
+    case SouthWestGravity:
+    case SouthEastGravity:
+	return -w->input.bottom;
+    }
+
+    return 0;
+}
+
 static Bool
 decorWindowUpdate (CompWindow *w,
-		   Bool	      move)
+		   Bool	      allowDecoration)
 {
     WindowDecoration *wd;
     Decoration	     *old, *decor = NULL;
+    Bool	     decorate = FALSE;
+    CompMatch	     *match;
+    int		     moveDx, moveDy;
+    int		     oldShiftX = 0;
+    int		     oldShiftY  = 0;
 
+    DECOR_DISPLAY (w->screen->display);
     DECOR_SCREEN (w->screen);
     DECOR_WINDOW (w);
 
     wd = dw->wd;
     old = (wd) ? wd->decor : NULL;
 
-    if (dw->decor && decorCheckSize (w, dw->decor))
-    {
-	if (w->type != CompWindowTypeFullscreenMask)
-	    decor = dw->decor;
+    switch (w->type) {
+    case CompWindowTypeDialogMask:
+    case CompWindowTypeModalDialogMask:
+    case CompWindowTypeUtilMask:
+    case CompWindowTypeNormalMask:
+	if (w->mwmDecor & (MwmDecorAll | MwmDecorTitle))
+	    decorate = TRUE;
+    default:
+	break;
     }
-    else
+
+    if (w->attrib.override_redirect)
+	decorate = FALSE;
+
+    if (decorate)
     {
-	if (w->attrib.override_redirect)
+	match = &dd->opt[DECOR_DISPLAY_OPTION_DECOR_MATCH].value.match;
+	if (!matchEval (match, w))
+	    decorate = FALSE;
+    }
+
+    if (decorate)
+    {
+	if (dw->decor && decorCheckSize (w, dw->decor))
 	{
-	    if (w->region->numRects == 1 && !w->alpha)
-		decor = ds->decor[DECOR_BARE];
+	    decor = dw->decor;
 	}
 	else
 	{
-	    switch (w->type) {
-	    case CompWindowTypeDialogMask:
-	    case CompWindowTypeModalDialogMask:
-	    case CompWindowTypeUtilMask:
-	    case CompWindowTypeNormalMask:
-		if (w->mwmDecor & (MwmDecorAll | MwmDecorTitle))
-		{
-		    if (w->id == w->screen->display->activeWindow)
-			decor = ds->decor[DECOR_ACTIVE];
-		    else
-			decor = ds->decor[DECOR_NORMAL];
-
-		    break;
-		}
-		/* fall-through */
-	    default:
-		if (w->region->numRects == 1 && !w->alpha)
-		    decor = ds->decor[DECOR_BARE];
-
-		/* no decoration on windows with below state */
-		if (w->state & CompWindowStateBelowMask)
-		    decor = NULL;
-
-		break;
-	    }
+	    if (w->id == w->screen->display->activeWindow)
+		decor = ds->decor[DECOR_ACTIVE];
+	    else
+		decor = ds->decor[DECOR_NORMAL];
 	}
-
-	if (decor)
+    }
+    else
+    {
+	match = &dd->opt[DECOR_DISPLAY_OPTION_SHADOW_MATCH].value.match;
+	if (matchEval (match, w))
 	{
-	    if (!decorCheckSize (w, decor))
+	    if (w->region->numRects == 1 && !w->alpha)
+		decor = ds->decor[DECOR_BARE];
+
+	    /* no decoration on windows with below state */
+	    if (w->state & CompWindowStateBelowMask)
 		decor = NULL;
+
+	    if (decor)
+	    {
+		if (!decorCheckSize (w, decor))
+		    decor = NULL;
+	    }
 	}
     }
 
-    if (!ds->dmWin)
+    if (!ds->dmWin || !allowDecoration)
 	decor = NULL;
 
     if (decor == old)
 	return FALSE;
 
+    damageWindowOutputExtents (w);
+
     if (old)
     {
-	damageWindowOutputExtents (w);
+	oldShiftX = decorWindowShiftX (w);
+	oldShiftY = decorWindowShiftY (w);
+
 	destroyWindowDecoration (w->screen, wd);
     }
 
@@ -936,18 +792,44 @@ decorWindowUpdate (CompWindow *w,
 	    return FALSE;
 
 	if ((w->state & MAXIMIZE_STATE) == MAXIMIZE_STATE)
-	    setWindowFrameExtents (w, &decor->maxInput, &decor->output);
+	    setWindowFrameExtents (w, &decor->maxInput);
 	else
-	    setWindowFrameExtents (w, &decor->input, &decor->output);
+	    setWindowFrameExtents (w, &decor->input);
 
-	if (!move)
-	    damageWindowOutputExtents (w);
+	moveDx = decorWindowShiftX (w) - oldShiftX;
+	moveDy = decorWindowShiftY (w) - oldShiftY;
 
+	updateWindowOutputExtents (w);
+	damageWindowOutputExtents (w);
 	updateWindowDecorationScale (w);
     }
     else
     {
 	dw->wd = NULL;
+
+	moveDx = -oldShiftX;
+	moveDy = -oldShiftY;
+    }
+
+    if (!w->attrib.override_redirect && (moveDx || moveDy))
+    {
+	XWindowChanges xwc;
+	unsigned int   mask = CWX | CWY;
+
+	xwc.x = w->serverX + moveDx;
+	xwc.y = w->serverY + moveDy;
+
+	if (w->state & CompWindowStateFullscreenMask)
+	    mask &= ~(CWX | CWY);
+
+	if (w->state & CompWindowStateMaximizedHorzMask)
+	    mask &= ~CWX;
+
+	if (w->state & CompWindowStateMaximizedVertMask)
+	    mask &= ~CWY;
+
+	if (mask)
+	    configureXWindow (w, mask, &xwc);
     }
 
     return TRUE;
@@ -1035,16 +917,12 @@ static void
 decorHandleEvent (CompDisplay *d,
 		  XEvent      *event)
 {
-    Window     activeWindow = 0;
+    Window     activeWindow = d->activeWindow;
     CompWindow *w;
 
     DECOR_DISPLAY (d);
 
     switch (event->type) {
-    case PropertyNotify:
-	if (event->xproperty.atom == d->winActiveAtom)
-	    activeWindow = d->activeWindow;
-	break;
     case DestroyNotify:
 	w = findWindowAtDisplay (d, event->xdestroywindow.window);
 	if (w)
@@ -1054,6 +932,12 @@ decorHandleEvent (CompDisplay *d,
 	    if (w->id == ds->dmWin)
 		decorCheckForDmOnScreen (w->screen, TRUE);
 	}
+	break;
+    case MapRequest:
+	w = findWindowAtDisplay (d, event->xmaprequest.window);
+	if (w)
+	    decorWindowUpdate (w, TRUE);
+	break;
     default:
 	if (event->type == d->damageEvent + XDamageNotify)
 	{
@@ -1076,9 +960,6 @@ decorHandleEvent (CompDisplay *d,
 
 			for (w = s->windows; w; w = w->next)
 			{
-			    if (w->attrib.override_redirect)
-				continue;
-
 			    if (w->shaded || w->mapNum)
 			    {
 				dw = GET_DECOR_WINDOW (w, ds);
@@ -1099,22 +980,20 @@ decorHandleEvent (CompDisplay *d,
     (*d->handleEvent) (d, event);
     WRAP (dd, d, handleEvent, decorHandleEvent);
 
+    if (d->activeWindow != activeWindow)
+    {
+	w = findWindowAtDisplay (d, activeWindow);
+	if (w)
+	    decorWindowUpdate (w, TRUE);
+
+	w = findWindowAtDisplay (d, d->activeWindow);
+	if (w)
+	    decorWindowUpdate (w, TRUE);
+    }
+
     switch (event->type) {
     case PropertyNotify:
-	if (event->xproperty.atom == d->winActiveAtom)
-	{
-	    if (d->activeWindow != activeWindow)
-	    {
-		w = findWindowAtDisplay (d, activeWindow);
-		if (w)
-		    decorWindowUpdate (w, FALSE);
-
-		w = findWindowAtDisplay (d, d->activeWindow);
-		if (w)
-		    decorWindowUpdate (w, FALSE);
-	    }
-	}
-	else if (event->xproperty.atom == dd->winDecorAtom)
+	if (event->xproperty.atom == dd->winDecorAtom)
 	{
 	    w = findWindowAtDisplay (d, event->xproperty.window);
 	    if (w)
@@ -1165,11 +1044,6 @@ decorHandleEvent (CompDisplay *d,
 	    }
 	}
 	break;
-    case MapRequest:
-	w = findWindowAtDisplay (d, event->xmaprequest.window);
-	if (w)
-	    decorWindowUpdate (w, TRUE);
-	break;
     default:
 	if (d->shapeExtension && event->type == d->shapeEvent + ShapeNotify)
 	{
@@ -1191,7 +1065,7 @@ decorDamageWindowRect (CompWindow *w,
     DECOR_SCREEN (w->screen);
 
     if (initial)
-	decorWindowUpdate (w, FALSE);
+	decorWindowUpdate (w, TRUE);
 
     UNWRAP (ds, w->screen, damageWindowRect);
     status = (*w->screen->damageWindowRect) (w, initial, rect);
@@ -1200,6 +1074,104 @@ decorDamageWindowRect (CompWindow *w,
     return status;
 }
 
+static void
+decorGetOutputExtentsForWindow (CompWindow	  *w,
+				CompWindowExtents *output)
+{
+    DECOR_SCREEN (w->screen);
+    DECOR_WINDOW (w);
+
+    UNWRAP (ds, w->screen, getOutputExtentsForWindow);
+    (*w->screen->getOutputExtentsForWindow) (w, output);
+    WRAP (ds, w->screen, getOutputExtentsForWindow,
+	  decorGetOutputExtentsForWindow);
+
+    if (dw->wd)
+    {
+	CompWindowExtents *e = &dw->wd->decor->output;
+
+	if (e->left > output->left)
+	    output->left = e->left;
+	if (e->right > output->right)
+	    output->right = e->right;
+	if (e->top > output->top)
+	    output->top = e->top;
+	if (e->bottom > output->bottom)
+	    output->bottom = e->bottom;
+    }
+}
+
+static CompOption *
+decorGetDisplayOptions (CompPlugin  *plugin,
+			CompDisplay *display,
+			int	    *count)
+{
+    DECOR_DISPLAY (display);
+
+    *count = NUM_OPTIONS (dd);
+    return dd->opt;
+}
+
+static Bool
+decorSetDisplayOption (CompPlugin      *plugin,
+		       CompDisplay     *display,
+		       char	       *name,
+		       CompOptionValue *value)
+{
+    CompOption *o;
+    int	       index;
+
+    DECOR_DISPLAY (display);
+
+    o = compFindOption (dd->opt, NUM_OPTIONS (dd), name, &index);
+    if (!o)
+	return FALSE;
+
+    switch (index) {
+    case DECOR_DISPLAY_OPTION_COMMAND:
+	if (compSetStringOption (o, value))
+	{
+	    if (display->screens && *o->value.s != '\0')
+	    {
+		DECOR_SCREEN (display->screens);
+
+		/* run decorator command if no decorator is present on
+		   first screen */
+		if (!ds->dmWin)
+		{
+		    if (fork () == 0)
+		    {
+			putenv (display->displayString);
+			execl ("/bin/sh", "/bin/sh", "-c", o->value.s, NULL);
+			exit (0);
+		    }
+		}
+	    }
+
+	    return TRUE;
+	}
+	break;
+    case DECOR_DISPLAY_OPTION_DECOR_MATCH:
+    case DECOR_DISPLAY_OPTION_SHADOW_MATCH:
+	if (compSetMatchOption (o, value))
+	{
+	    CompScreen *s;
+	    CompWindow *w;
+
+	    for (s = display->screens; s; s = s->next)
+		for (w = s->windows; w; w = w->next)
+		    decorWindowUpdate (w, TRUE);
+	}
+	break;
+    default:
+	if (compSetOption (o, value))
+	    return TRUE;
+	break;
+    }
+
+    return FALSE;
+}
+
 static void
 decorWindowMoveNotify (CompWindow *w,
 		       int	  dx,
@@ -1231,15 +1203,19 @@ decorWindowMoveNotify (CompWindow *w,
 }
 
 static void
-decorWindowResizeNotify (CompWindow *w)
+decorWindowResizeNotify (CompWindow *w,
+			 int	    dx,
+			 int	    dy,
+			 int	    dwidth,
+			 int	    dheight)
 {
     DECOR_SCREEN (w->screen);
 
-    if (!decorWindowUpdate (w, FALSE))
+    if (!decorWindowUpdate (w, TRUE))
 	updateWindowDecorationScale (w);
 
     UNWRAP (ds, w->screen, windowResizeNotify);
-    (*w->screen->windowResizeNotify) (w);
+    (*w->screen->windowResizeNotify) (w, dx, dy, dwidth, dheight);
     WRAP (ds, w->screen, windowResizeNotify, decorWindowResizeNotify);
 }
 
@@ -1249,14 +1225,15 @@ decorWindowStateChangeNotify (CompWindow *w)
     DECOR_SCREEN (w->screen);
     DECOR_WINDOW (w);
 
-    if (dw->wd && dw->wd->decor)
+    if (!decorWindowUpdate (w, TRUE))
     {
-	Decoration *decor = dw->wd->decor;
-
-	if ((w->state & MAXIMIZE_STATE) == MAXIMIZE_STATE)
-	    setWindowFrameExtents (w, &decor->maxInput, &decor->output);
-	else
-	    setWindowFrameExtents (w, &decor->input, &decor->output);
+	if (dw->decor)
+	{
+	    if ((w->state & MAXIMIZE_STATE) == MAXIMIZE_STATE)
+		setWindowFrameExtents (w, &dw->decor->maxInput);
+	    else
+		setWindowFrameExtents (w, &dw->decor->input);
+	}
     }
 
     UNWRAP (ds, w->screen, windowStateChangeNotify);
@@ -1264,6 +1241,44 @@ decorWindowStateChangeNotify (CompWindow *w)
     WRAP (ds, w->screen, windowStateChangeNotify, decorWindowStateChangeNotify);
 }
 
+static void
+decorMatchPropertyChanged (CompDisplay *d,
+			   CompWindow  *w)
+{
+    DECOR_DISPLAY (d);
+
+    decorWindowUpdate (w, TRUE);
+
+    UNWRAP (dd, d, matchPropertyChanged);
+    (*d->matchPropertyChanged) (d, w);
+    WRAP (dd, d, matchPropertyChanged, decorMatchPropertyChanged);
+}
+
+static void
+decorWindowAddNotify (CompWindow *w)
+{
+    DECOR_SCREEN (w->screen);
+
+    if (w->shaded || w->attrib.map_state == IsViewable)
+	decorWindowUpdate (w, TRUE);
+
+    UNWRAP (ds, w->screen, windowAddNotify);
+    (*w->screen->windowAddNotify) (w);
+    WRAP (ds, w->screen, windowAddNotify, decorWindowAddNotify);
+}
+
+static const CompMetadataOptionInfo decorDisplayOptionInfo[] = {
+    { "shadow_radius", "float", "<min>0.0</min><max>48.0</max>", 0, 0 },
+    { "shadow_opacity", "float", "<min>0.0</min>", 0, 0 },
+    { "shadow_color", "color", 0, 0, 0 },
+    { "shadow_x_offset", "int", "<min>-16</min><max>16</max>", 0, 0 },
+    { "shadow_y_offset", "int", "<min>-16</min><max>16</max>", 0, 0 },
+    { "command", "string", 0, 0, 0 },
+    { "mipmap", "bool", 0, 0, 0 },
+    { "decoration_match", "match", 0, 0, 0 },
+    { "shadow_match", "match", 0, 0, 0 }
+};
+
 static Bool
 decorInitDisplay (CompPlugin  *p,
 		  CompDisplay *d)
@@ -1274,9 +1289,20 @@ decorInitDisplay (CompPlugin  *p,
     if (!dd)
 	return FALSE;
 
+    if (!compInitDisplayOptionsFromMetadata (d,
+					     &decorMetadata,
+					     decorDisplayOptionInfo,
+					     dd->opt,
+					     DECOR_DISPLAY_OPTION_NUM))
+    {
+	free (dd);
+	return FALSE;
+    }
+
     dd->screenPrivateIndex = allocateScreenPrivateIndex (d);
     if (dd->screenPrivateIndex < 0)
     {
+	compFiniDisplayOptions (d, dd->opt, DECOR_DISPLAY_OPTION_NUM);
 	free (dd);
 	return FALSE;
     }
@@ -1293,9 +1319,8 @@ decorInitDisplay (CompPlugin  *p,
     dd->decorAtom[DECOR_ACTIVE] =
 	XInternAtom (d->display, "_NET_WINDOW_DECOR_ACTIVE", 0);
 
-    decorDisplayInitOptions (dd);
-
     WRAP (dd, d, handleEvent, decorHandleEvent);
+    WRAP (dd, d, matchPropertyChanged, decorMatchPropertyChanged);
 
     d->privates[displayPrivateIndex].ptr = dd;
 
@@ -1311,6 +1336,9 @@ decorFiniDisplay (CompPlugin  *p,
     freeScreenPrivateIndex (d, dd->screenPrivateIndex);
 
     UNWRAP (dd, d, handleEvent);
+    UNWRAP (dd, d, matchPropertyChanged);
+
+    compFiniDisplayOptions (d, dd->opt, DECOR_DISPLAY_OPTION_NUM);
 
     free (dd);
 }
@@ -1338,8 +1366,10 @@ decorInitScreen (CompPlugin *p,
 
     ds->dmWin = None;
 
-    WRAP (ds, s, paintWindow, decorPaintWindow);
+    WRAP (ds, s, drawWindow, decorDrawWindow);
     WRAP (ds, s, damageWindowRect, decorDamageWindowRect);
+    WRAP (ds, s, getOutputExtentsForWindow, decorGetOutputExtentsForWindow);
+    WRAP (ds, s, windowAddNotify, decorWindowAddNotify);
     WRAP (ds, s, windowMoveNotify, decorWindowMoveNotify);
     WRAP (ds, s, windowResizeNotify, decorWindowResizeNotify);
     WRAP (ds, s, windowStateChangeNotify, decorWindowStateChangeNotify);
@@ -1363,8 +1393,10 @@ decorFiniScreen (CompPlugin *p,
 	if (ds->decor[i])
 	    decorReleaseDecoration (s, ds->decor[i]);
 
-    UNWRAP (ds, s, paintWindow);
+    UNWRAP (ds, s, drawWindow);
     UNWRAP (ds, s, damageWindowRect);
+    UNWRAP (ds, s, getOutputExtentsForWindow);
+    UNWRAP (ds, s, windowAddNotify);
     UNWRAP (ds, s, windowMoveNotify);
     UNWRAP (ds, s, windowResizeNotify);
     UNWRAP (ds, s, windowStateChangeNotify);
@@ -1392,8 +1424,8 @@ decorInitWindow (CompPlugin *p,
     if (!w->attrib.override_redirect)
 	decorWindowUpdateDecoration (w);
 
-    if (w->shaded || w->attrib.map_state == IsViewable)
-	decorWindowUpdate (w, FALSE);
+    if (w->added && (w->shaded || w->attrib.map_state == IsViewable))
+	decorWindowUpdate (w, TRUE);
 
     return TRUE;
 }
@@ -1404,6 +1436,9 @@ decorFiniWindow (CompPlugin *p,
 {
     DECOR_WINDOW (w);
 
+    if (!w->destroyed)
+	decorWindowUpdate (w, FALSE);
+
     if (dw->wd)
 	destroyWindowDecoration (w->screen, dw->wd);
 
@@ -1416,9 +1451,21 @@ decorFiniWindow (CompPlugin *p,
 static Bool
 decorInit (CompPlugin *p)
 {
+    if (!compInitPluginMetadataFromInfo (&decorMetadata,
+					 p->vTable->name,
+					 decorDisplayOptionInfo,
+					 DECOR_DISPLAY_OPTION_NUM,
+					 0, 0))
+	return FALSE;
+
     displayPrivateIndex = allocateDisplayPrivateIndex ();
     if (displayPrivateIndex < 0)
+    {
+	compFiniMetadata (&decorMetadata);
 	return FALSE;
+    }
+
+    compAddMetadataFromFile (&decorMetadata, p->vTable->name);
 
     return TRUE;
 }
@@ -1426,8 +1473,8 @@ decorInit (CompPlugin *p)
 static void
 decorFini (CompPlugin *p)
 {
-    if (displayPrivateIndex >= 0)
-	freeDisplayPrivateIndex (displayPrivateIndex);
+    freeDisplayPrivateIndex (displayPrivateIndex);
+    compFiniMetadata (&decorMetadata);
 }
 
 static int
@@ -1437,22 +1484,16 @@ decorGetVersion (CompPlugin *plugin,
     return ABIVERSION;
 }
 
-CompPluginDep decorDeps[] = {
-    { CompPluginRuleBefore, "wobbly" },
-    { CompPluginRuleBefore, "fade" },
-    { CompPluginRuleBefore, "cube" },
-    { CompPluginRuleBefore, "scale" }
-};
-
-CompPluginFeature decorFeatures[] = {
-    { "decorations" }
-};
+static CompMetadata *
+decorGetMetadata (CompPlugin *plugin)
+{
+    return &decorMetadata;
+}
 
 static CompPluginVTable decorVTable = {
     "decoration",
-    N_("Window Decoration"),
-    N_("Window decorations"),
     decorGetVersion,
+    decorGetMetadata,
     decorInit,
     decorFini,
     decorInitDisplay,
@@ -1464,11 +1505,7 @@ static CompPluginVTable decorVTable = {
     decorGetDisplayOptions,
     decorSetDisplayOption,
     0, /* GetScreenOptions */
-    0, /* SetScreenOption */
-    decorDeps,
-    sizeof (decorDeps) / sizeof (decorDeps[0]),
-    decorFeatures,
-    sizeof (decorFeatures) / sizeof (decorFeatures[0])
+    0  /* SetScreenOption */
 };
 
 CompPluginVTable *
