@@ -35,21 +35,20 @@
 #include <glib/gprintf.h>
 #include <gconf/gconf-client.h>
 
+static CompMetadata gconfMetadata;
+
 #define APP_NAME "/apps/compiz"
 
 /* From gconf-internal.h. Bleah. */
 int gconf_value_compare (const GConfValue *value_a,
 			 const GConfValue *value_b);
 
-#define KEY_CHANGE_TIMEOUT 250
-
 static int displayPrivateIndex;
 
 typedef struct _GConfDisplay {
     int screenPrivateIndex;
 
-    GConfClient       *client;
-    CompTimeoutHandle timeoutHandle;
+    GConfClient *client;
 
     InitPluginForDisplayProc      initPluginForDisplay;
     SetDisplayOptionProc	  setDisplayOption;
@@ -104,6 +103,8 @@ gconfTypeFromCompType (CompOptionType type)
 	return GCONF_VALUE_STRING;
     case CompOptionTypeAction:
 	return GCONF_VALUE_STRING;
+    case CompOptionTypeMatch:
+	return GCONF_VALUE_STRING;
     case CompOptionTypeList:
 	return GCONF_VALUE_LIST;
     default:
@@ -140,6 +141,14 @@ gconfSetValue (CompDisplay     *d,
 
 	free (color);
     } break;
+    case CompOptionTypeMatch: {
+	gchar *match;
+
+	match = matchToString (&value->match);
+	gconf_value_set_string (gvalue, match);
+
+	free (match);
+    } break;
     default:
 	break;
     }
@@ -201,6 +210,7 @@ gconfSetOption (CompDisplay *d,
     case CompOptionTypeFloat:
     case CompOptionTypeString:
     case CompOptionTypeColor:
+    case CompOptionTypeMatch:
 	existingValue = gconf_client_get (gd->client, key, NULL);
 	gvalue = gconf_value_new (gconfTypeFromCompType (o->type));
 	gconfSetValue (d, &o->value, o->type, gvalue);
@@ -382,6 +392,18 @@ gconfGetValue (CompDisplay     *d,
 	if (stringToColor (color, value->c))
 	    return TRUE;
     }
+    else if (type         == CompOptionTypeMatch &&
+	     gvalue->type == GCONF_VALUE_STRING)
+    {
+	const gchar *match;
+
+	match = gconf_value_get_string (gvalue);
+
+	matchInit (&value->match);
+	matchAddFromString (&value->match, match);
+
+	return TRUE;
+    }
 
     return FALSE;
 }
@@ -455,12 +477,12 @@ gconfGetOptionValue (CompDisplay *d,
 	    if (s)
 	    {
 		if (p->vTable->getScreenOptions)
-		    option = (*p->vTable->getScreenOptions) (s, &nOption);
+		    option = (*p->vTable->getScreenOptions) (p, s, &nOption);
 	    }
 	    else
 	    {
 		if (p->vTable->getDisplayOptions)
-		    option = (*p->vTable->getDisplayOptions) (d, &nOption);
+		    option = (*p->vTable->getDisplayOptions) (p, d, &nOption);
 	    }
 	}
     }
@@ -739,9 +761,21 @@ gconfGetOptionValue (CompDisplay *d,
 
 		    if (o->type == CompOptionTypeList)
 		    {
+			if (o->value.list.type == CompOptionTypeMatch)
+			{
+			    int i;
+
+			    for (i = 0; i < value.list.nValue; i++)
+				matchFini (&value.list.value[i].match);
+			}
+
 			if (value.list.value)
 			    free (value.list.value);
 		    }
+		    else if (o->type == CompOptionTypeMatch)
+		    {
+			matchFini (&value.match);
+		    }
 		}
 	    }
 
@@ -829,7 +863,7 @@ gconfSetDisplayOptionForPlugin (CompDisplay     *d,
 	    CompOption *option;
 	    int	       nOption;
 
-	    option = (*p->vTable->getDisplayOptions) (d, &nOption);
+	    option = (*p->vTable->getDisplayOptions) (p, d, &nOption);
 	    gconfSetOption (d, compFindOption (option, nOption, name, 0),
 			    "allscreens", plugin);
 	}
@@ -896,7 +930,7 @@ gconfSetScreenOptionForPlugin (CompScreen      *s,
 
 	    screen = g_strdup_printf ("screen%d", s->screenNum);
 
-	    option = (*p->vTable->getScreenOptions) (s, &nOption);
+	    option = (*p->vTable->getScreenOptions) (p, s, &nOption);
 	    gconfSetOption (s->display,
 			    compFindOption (option, nOption, name, 0),
 			    screen, plugin);
@@ -925,7 +959,7 @@ gconfInitPluginForDisplay (CompPlugin  *p,
 	CompOption *option;
 	int	   nOption;
 
-	option = (*p->vTable->getDisplayOptions) (d, &nOption);
+	option = (*p->vTable->getDisplayOptions) (p, d, &nOption);
 	while (nOption--)
 	    gconfInitOption (d, option++, "allscreens", p->vTable->name);
     }
@@ -953,7 +987,7 @@ gconfInitPluginForScreen (CompPlugin *p,
 
 	screen = g_strdup_printf ("screen%d", s->screenNum);
 
-	option = (*p->vTable->getScreenOptions) (s, &nOption);
+	option = (*p->vTable->getScreenOptions) (p, s, &nOption);
 	while (nOption--)
 	    gconfInitOption (s->display, option++, screen, p->vTable->name);
 
@@ -988,12 +1022,26 @@ gconfKeyChanged (GConfClient *client,
 	g_free (key);
 }
 
-static Bool
-gconfTimeout (void *closure)
+static void
+gconfSendGLibNotify (CompDisplay *d)
 {
-    while (g_main_pending ()) g_main_iteration (FALSE);
+    Display *dpy = d->display;
+    XEvent  xev;
 
-    return TRUE;
+    xev.xclient.type    = ClientMessage;
+    xev.xclient.display = dpy;
+    xev.xclient.format  = 32;
+
+    xev.xclient.message_type = XInternAtom (dpy, "_COMPIZ_GLIB_NOTIFY", 0);
+    xev.xclient.window	     = d->screens->root;
+
+    memset (xev.xclient.data.l, 0, sizeof (xev.xclient.data.l));
+
+    XSendEvent (dpy,
+		d->screens->root,
+		FALSE,
+		SubstructureRedirectMask | SubstructureNotifyMask,
+		&xev);
 }
 
 static Bool
@@ -1035,7 +1083,7 @@ gconfInitDisplay (CompPlugin  *p,
     gconf_client_notify_add (gd->client, APP_NAME, gconfKeyChanged, d,
 			     NULL, NULL);
 
-    gd->timeoutHandle = compAddTimeout (KEY_CHANGE_TIMEOUT, gconfTimeout, 0);
+    gconfSendGLibNotify (d);
 
     return TRUE;
 }
@@ -1046,8 +1094,6 @@ gconfFiniDisplay (CompPlugin  *p,
 {
     GCONF_DISPLAY (d);
 
-    compRemoveTimeout (gd->timeoutHandle);
-
     g_object_unref (gd->client);
 
     UNWRAP (gd, d, initPluginForDisplay);
@@ -1107,9 +1153,18 @@ gconfFiniScreen (CompPlugin *p,
 static Bool
 gconfInit (CompPlugin *p)
 {
+    if (!compInitPluginMetadataFromInfo (&gconfMetadata, p->vTable->name,
+					 0, 0, 0, 0))
+	return FALSE;
+
     displayPrivateIndex = allocateDisplayPrivateIndex ();
     if (displayPrivateIndex < 0)
+    {
+	compFiniMetadata (&gconfMetadata);
 	return FALSE;
+    }
+
+    compAddMetadataFromFile (&gconfMetadata, p->vTable->name);
 
     return TRUE;
 }
@@ -1117,8 +1172,8 @@ gconfInit (CompPlugin *p)
 static void
 gconfFini (CompPlugin *p)
 {
-    if (displayPrivateIndex >= 0)
-	freeDisplayPrivateIndex (displayPrivateIndex);
+    freeDisplayPrivateIndex (displayPrivateIndex);
+    compFiniMetadata (&gconfMetadata);
 }
 
 static int
@@ -1128,19 +1183,16 @@ gconfGetVersion (CompPlugin *plugin,
     return ABIVERSION;
 }
 
-CompPluginDep gconfDeps[] = {
-    { CompPluginRuleBefore, "decoration" },
-    { CompPluginRuleBefore, "wobbly" },
-    { CompPluginRuleBefore, "fade" },
-    { CompPluginRuleBefore, "cube" },
-    { CompPluginRuleBefore, "scale" }
-};
+static CompMetadata *
+gconfGetMetadata (CompPlugin *plugin)
+{
+    return &gconfMetadata;
+}
 
 CompPluginVTable gconfVTable = {
     "gconf",
-    "GConf",
-    "GConf Control Backend",
     gconfGetVersion,
+    gconfGetMetadata,
     gconfInit,
     gconfFini,
     gconfInitDisplay,
@@ -1152,11 +1204,7 @@ CompPluginVTable gconfVTable = {
     0, /* GetDisplayOptions */
     0, /* SetDisplayOption */
     0, /* GetScreenOptions */
-    0, /* SetScreenOption */
-    gconfDeps,
-    sizeof (gconfDeps) / sizeof (gconfDeps[0]),
-    0, /* Features */
-    0  /* nFeatures */
+    0  /* SetScreenOption */
 };
 
 CompPluginVTable *
