@@ -22,27 +22,47 @@
 
 #include <math.h>
 #include <stdlib.h>
+#include <string.h>
 
 #include <compiz.h>
 
 #include <glib.h>
 
-#define PLACE_WORKAROUND_DEFAULT TRUE
+static CompMetadata placeMetadata;
+
+#define PLACE_MODE_CASCADE  0
+#define PLACE_MODE_CENTERED 1
+#define PLACE_MODE_SMART    2
+#define PLACE_MODE_MAXIMIZE 3
+#define PLACE_MODE_RANDOM   4
+#define PLACE_MODE_LAST     PLACE_MODE_RANDOM
+
+/* overlap types */
+#define NONE    0
+#define H_WRONG -1
+#define W_WRONG -2
 
 static int displayPrivateIndex;
 
 typedef struct _PlaceDisplay {
     int		    screenPrivateIndex;
-    HandleEventProc handleEvent;
 } PlaceDisplay;
 
-#define PLACE_SCREEN_OPTION_WORKAROUND 0
-#define PLACE_SCREEN_OPTION_NUM        1
+#define PLACE_SCREEN_OPTION_WORKAROUND        0
+#define PLACE_SCREEN_OPTION_MODE              1
+#define PLACE_SCREEN_OPTION_POSITION_MATCHES  2
+#define PLACE_SCREEN_OPTION_POSITION_X_VALUES 3
+#define PLACE_SCREEN_OPTION_POSITION_Y_VALUES 4
+#define PLACE_SCREEN_OPTION_VIEWPORT_MATCHES  5
+#define PLACE_SCREEN_OPTION_VIEWPORT_X_VALUES 6
+#define PLACE_SCREEN_OPTION_VIEWPORT_Y_VALUES 7
+#define PLACE_SCREEN_OPTION_NUM               8
 
 typedef struct _PlaceScreen {
     CompOption opt[PLACE_SCREEN_OPTION_NUM];
 
-    DamageWindowRectProc damageWindowRect;
+    PlaceWindowProc placeWindow;
+
 } PlaceScreen;
 
 #define GET_PLACE_DISPLAY(d)				      \
@@ -59,8 +79,69 @@ typedef struct _PlaceScreen {
 
 #define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))
 
+static Bool
+placeMatchXYValue (CompWindow *w,
+		   CompOption *matches,
+		   CompOption *xValues,
+		   CompOption *yValues,
+		   int	      *x,
+		   int	      *y)
+{
+    int i, min;
+
+    if (w->type & CompWindowTypeDesktopMask)
+	return FALSE;
+
+    min = MIN (matches->value.list.nValue, xValues->value.list.nValue);
+    min = MIN (min, yValues->value.list.nValue);
+
+    for (i = 0; i < min; i++)
+    {
+	if (matchEval (&matches->value.list.value[i].match, w))
+	{
+	    *x = xValues->value.list.value[i].i;
+	    *y = yValues->value.list.value[i].i;
+
+	    return TRUE;
+	}
+    }
+
+    return FALSE;
+}
+
+static Bool
+placeMatchPosition (CompWindow *w,
+		    int	       *x,
+		    int	       *y)
+{
+    PLACE_SCREEN (w->screen);
+
+    return placeMatchXYValue (w,
+			      &ps->opt[PLACE_SCREEN_OPTION_POSITION_MATCHES],
+			      &ps->opt[PLACE_SCREEN_OPTION_POSITION_X_VALUES],
+			      &ps->opt[PLACE_SCREEN_OPTION_POSITION_Y_VALUES],
+			      x,
+			      y);
+}
+
+static Bool
+placeMatchViewport (CompWindow *w,
+		    int	       *x,
+		    int	       *y)
+{
+    PLACE_SCREEN (w->screen);
+
+    return placeMatchXYValue (w,
+			      &ps->opt[PLACE_SCREEN_OPTION_VIEWPORT_MATCHES],
+			      &ps->opt[PLACE_SCREEN_OPTION_VIEWPORT_X_VALUES],
+			      &ps->opt[PLACE_SCREEN_OPTION_VIEWPORT_Y_VALUES],
+			      x,
+			      y);
+}
+
 static CompOption *
-placeGetScreenOptions (CompScreen *screen,
+placeGetScreenOptions (CompPlugin *plugin,
+		       CompScreen *screen,
 		       int	  *count)
 {
     PLACE_SCREEN (screen);
@@ -70,7 +151,8 @@ placeGetScreenOptions (CompScreen *screen,
 }
 
 static Bool
-placeSetScreenOption (CompScreen      *screen,
+placeSetScreenOption (CompPlugin      *plugin,
+		      CompScreen      *screen,
 		      char	      *name,
 		      CompOptionValue *value)
 {
@@ -84,29 +166,31 @@ placeSetScreenOption (CompScreen      *screen,
 	return FALSE;
 
     switch (index) {
-    case PLACE_SCREEN_OPTION_WORKAROUND:
-	if (compSetBoolOption (o, value))
+    case PLACE_SCREEN_OPTION_MODE:
+	if (compSetIntOption (o, value))
 	    return TRUE;
+	break;
+    case PLACE_SCREEN_OPTION_POSITION_MATCHES:
+    case PLACE_SCREEN_OPTION_VIEWPORT_MATCHES:
+	if (compSetOptionList (o, value))
+	{
+	    int i;
+
+	    for (i = 0; i < o->value.list.nValue; i++)
+		matchUpdate (screen->display, &o->value.list.value[i].match);
+
+	    return TRUE;
+       }
+       break;
     default:
+	if (compSetOption (o, value))
+	    return TRUE;
 	break;
     }
 
     return FALSE;
 }
 
-static void
-placeScreenInitOptions (PlaceScreen *ps)
-{
-    CompOption *o;
-
-    o = &ps->opt[PLACE_SCREEN_OPTION_WORKAROUND];
-    o->name	 = "workarounds";
-    o->shortDesc = N_("Workarounds");
-    o->longDesc	 = N_("Window placement workarounds");
-    o->type	 = CompOptionTypeBool;
-    o->value.b	 = PLACE_WORKAROUND_DEFAULT;
-}
-
 typedef enum {
     PlaceLeft,
     PlaceRight,
@@ -199,6 +283,16 @@ get_window_height (CompWindow *window)
     return window->serverHeight + window->serverBorderWidth * 2;
 }
 
+static void
+get_outer_rect_of_window (CompWindow *w,
+			  XRectangle *r)
+{
+    r->x      = w->serverX - w->input.left;
+    r->y      = w->serverY - w->input.top;
+    r->width  = get_window_width (w)  + w->input.left + w->input.right;
+    r->height = get_window_height (w) + w->input.top  + w->input.bottom;
+}
+
 static void
 find_next_cascade (CompWindow *window,
 		   GList      *windows,
@@ -347,8 +441,8 @@ find_most_freespace (CompWindow *window,
 
     get_workarea_of_current_output_device (window->screen, &work_area);
 
-    getOuterRectOfWindow (focus_window, &avoid);
-    getOuterRectOfWindow (window, &outer);
+    get_outer_rect_of_window (focus_window, &avoid);
+    get_outer_rect_of_window (window, &outer);
 
     /* Find the areas of choosing the various sides of the focus window */
     max_width  = MIN (avoid.width, outer.width);
@@ -489,7 +583,7 @@ rectangle_overlaps_some_window (XRectangle *rect,
 	case CompWindowTypeUtilMask:
 	case CompWindowTypeToolbarMask:
 	case CompWindowTypeMenuMask:
-	    getOuterRectOfWindow (other, &other_rect);
+	    get_outer_rect_of_window (other, &other_rect);
 
 	    if (rectangleIntersect (rect, &other_rect, &dest))
 		return TRUE;
@@ -610,7 +704,7 @@ find_first_fit (CompWindow *window,
     right_sorted = g_list_sort (right_sorted, topmost_cmp);
     right_sorted = g_list_sort (right_sorted, leftmost_cmp);
 
-    getOuterRectOfWindow (window, &rect);
+    get_outer_rect_of_window (window, &rect);
 
     get_workarea_of_current_output_device (window->screen, &work_area);
 
@@ -639,7 +733,7 @@ find_first_fit (CompWindow *window,
 	CompWindow *w = tmp->data;
 	XRectangle outer_rect;
 
-	getOuterRectOfWindow (w, &outer_rect);
+	get_outer_rect_of_window (w, &outer_rect);
 
 	rect.x = outer_rect.x;
 	rect.y = outer_rect.y + outer_rect.height;
@@ -665,7 +759,7 @@ find_first_fit (CompWindow *window,
 	CompWindow *w = tmp->data;
 	XRectangle outer_rect;
 
-	getOuterRectOfWindow (w, &outer_rect);
+	get_outer_rect_of_window (w, &outer_rect);
 
 	rect.x = outer_rect.x + outer_rect.width;
 	rect.y = outer_rect.y;
@@ -692,11 +786,231 @@ out:
 }
 
 static void
-placeWindow (CompWindow *window,
-	     int        x,
-	     int        y,
-	     int        *new_x,
-	     int        *new_y)
+placeCentered (CompWindow *window,
+	       XRectangle *workarea,
+	       int	  *x,
+	       int	  *y)
+{
+    *x = workarea->x + (workarea->width - get_window_width (window)) / 2;
+    *y = workarea->y + (workarea->height - get_window_height (window)) / 2;
+}
+
+static void
+placeRandom (CompWindow *window,
+	     XRectangle *workarea,
+	     int	*x,
+	     int	*y)
+{
+    *x = workarea->x + rand () % (workarea->width - get_window_width (window));
+    *y = workarea->y + rand () % (workarea->height -
+				  get_window_height (window));
+}
+
+static void
+placeSmart (CompWindow *window,
+	    XRectangle *workarea,
+	    int        *x,
+	    int        *y)
+{
+    /*
+     * SmartPlacement by Cristian Tibirna (tibirna@kde.org)
+     * adapted for kwm (16-19jan98) and for kwin (16Nov1999) using (with
+     * permission) ideas from fvwm, authored by
+     * Anthony Martin (amartin@engr.csulb.edu).
+     * Xinerama supported added by Balaji Ramani (balaji@yablibli.com)
+     * with ideas from xfce.
+     * adapted for Compiz by Bellegarde Cedric (gnumdk(at)gmail.com)
+     */
+    CompWindow *wi;
+    long int overlap, minOverlap = 0;
+    int xOptimal, yOptimal;
+    int possible;
+
+    /* temp coords */
+    int cxl, cxr, cyt, cyb;
+    /* temp coords */
+    int  xl,  xr,  yt,  yb;
+    /* temp holder */
+    int basket;
+    /* CT lame flag. Don't like it. What else would do? */
+    Bool firstPass = TRUE;
+
+    /* get the maximum allowed windows space */
+    int xTmp = workarea->x;
+    int yTmp = workarea->y;
+
+    xOptimal = xTmp; yOptimal = yTmp;
+
+    /* client gabarit */
+    int ch = get_window_height (window) + window->input.left +
+	     window->input.right - 1;
+    int cw = get_window_width (window) + window->input.top +
+	     window->input.bottom - 1;
+
+    /* loop over possible positions */
+    do
+    {
+	/* test if enough room in x and y directions */
+	if (yTmp + ch > (workarea->y + workarea->height) &&
+	    ch < workarea->height)
+	    overlap = H_WRONG; /* this throws the algorithm to an exit */
+	else if (xTmp + cw > (workarea->x + workarea->width))
+	    overlap = W_WRONG;
+	else
+	{
+	    overlap = NONE; /* initialize */
+
+	    cxl = xTmp;
+	    cxr = xTmp + cw;
+	    cyt = yTmp;
+	    cyb = yTmp + ch;
+
+	    for (wi = window->screen->windows; wi; wi = wi->next)
+	    {
+		if (!wi->invisible &&
+		    wi != window &&
+		    !(wi->wmType & (CompWindowTypeDockMask |
+				    CompWindowTypeDesktopMask)))
+		{
+
+		    xl = wi->attrib.x - wi->input.left;
+		    yt = wi->attrib.y - wi->input.top;
+		    xr = xl + get_window_width (wi) + window->input.left
+			+ wi->input.right;
+		    yb = yt + get_window_height (wi) + window->input.top
+			+ wi->input.bottom;
+
+		    /* if windows overlap, calc the overall overlapping */
+		    if ((cxl < xr) && (cxr > xl) &&
+			(cyt < yb) && (cyb > yt))
+		    {
+			xl = MAX (cxl, xl); xr = MIN (cxr, xr);
+			yt = MAX (cyt, yt); yb = MIN (cyb, yb);
+			if (wi->state & CompWindowStateAboveMask)
+			    overlap += 16 * (xr - xl) * (yb - yt);
+			else if (wi->state & CompWindowStateBelowMask)
+			    overlap += 0;
+			else
+			    overlap += (xr - xl) * (yb - yt);
+		    }
+		}
+	    }
+	}
+
+	/* CT first time we get no overlap we stop */
+	if (overlap == NONE)
+	{
+	    xOptimal = xTmp;
+	    yOptimal = yTmp;
+	    break;
+	}
+
+	if (firstPass)
+	{
+	    firstPass = FALSE;
+	    minOverlap = overlap;
+	}
+	/* CT save the best position and the minimum overlap up to now */
+	else if (overlap >= NONE && overlap < minOverlap)
+	{
+	    minOverlap = overlap;
+	    xOptimal = xTmp;
+	    yOptimal = yTmp;
+	}
+
+	/* really need to loop? test if there's any overlap */
+	if (overlap > NONE)
+	{
+	    possible = workarea->x + workarea->width;
+
+	    if (possible - cw > xTmp) possible -= cw;
+
+	    /* compare to the position of each client on the same desk */
+	    for (wi = window->screen->windows; wi; wi = wi->next)
+	    {
+
+		if (!wi->invisible &&
+		    wi != window &&
+		    !(wi->wmType & (CompWindowTypeDockMask |
+				    CompWindowTypeDesktopMask)))
+		{
+
+		    xl = wi->attrib.x - wi->input.left;
+		    yt = wi->attrib.y - wi->input.top;
+		    xr = xl + get_window_width (wi) + wi->input.left
+			+ wi->input.right;
+		    yb = yt + get_window_height (wi) + wi->input.top
+			+ wi->input.bottom;
+
+		    /* if not enough room above or under the current
+		     * client determine the first non-overlapped x position
+		     */
+		    if ((yTmp < yb) && (yt < ch + yTmp))
+		    {
+			if ((xr > xTmp) && (possible > xr)) possible = xr;
+
+			basket = xl - cw;
+			if ((basket > xTmp) && (possible > basket))
+			    possible = basket;
+		    }
+		}
+	    }
+	    xTmp = possible;
+	}
+
+	/* else ==> not enough x dimension (overlap was wrong on horizontal) */
+	else if (overlap == W_WRONG)
+	{
+	    xTmp = workarea->x;
+	    possible = workarea->y + workarea->height;
+
+	    if (possible - ch > yTmp) possible -= ch;
+
+	    /* test the position of each window on the desk */
+	    for (wi = window->screen->windows; wi ; wi = wi->next)
+	    {
+		if (!wi->invisible &&
+		    wi != window &&
+		    !(wi->wmType & (CompWindowTypeDockMask |
+				    CompWindowTypeDesktopMask)))
+		{
+		    xl = wi->attrib.x - wi->input.left;
+		    yt = wi->attrib.y - wi->input.top;
+		    xr = xl + get_window_width (wi) + wi->input.left
+			+ wi->input.right;
+		    yb = yt + get_window_height (wi) + wi->input.top
+			+ wi->input.bottom;
+
+		    /* if not enough room to the left or right of the current
+		     * client determine the first non-overlapped y position
+		     */
+		    if ((yb > yTmp) && (possible > yb))
+			possible = yb;
+
+		    basket = yt - ch;
+		    if ((basket > yTmp) && (possible > basket))
+			possible = basket;
+		}
+	    }
+	    yTmp = possible;
+	}
+    }
+    while ((overlap != NONE) && (overlap != H_WRONG) && yTmp <
+	   (workarea->y + workarea->height));
+
+    if (ch >= workarea->height)
+	yOptimal = workarea->y;
+
+    *x = xOptimal + window->input.left;
+    *y = yOptimal + window->input.top;
+}
+
+static void
+placeWin (CompWindow *window,
+     	  int        x,
+	  int        y,
+	  int        *new_x,
+	  int        *new_y)
 {
     CompWindow *wi;
     GList      *windows;
@@ -916,13 +1230,34 @@ placeWindow (CompWindow *window,
     x = x0;
     y = y0;
 
-    if (find_first_fit (window, windows, x, y, &x, &y))
-	goto done_check_denied_focus;
+    if (!placeMatchPosition (window, &x, &y))
+    {
+	switch (ps->opt[PLACE_SCREEN_OPTION_MODE].value.i) {
+	case PLACE_MODE_CASCADE:
+	    if (find_first_fit (window, windows, x, y, &x, &y))
+		goto done_check_denied_focus;
 
-    /* if the window wasn't placed at the origin of screen,
-     * cascade it onto the current screen
-     */
-    find_next_cascade (window, windows, x, y, &x, &y);
+	    /* if the window wasn't placed at the origin of screen,
+	     * cascade it onto the current screen
+	     */
+	    find_next_cascade (window, windows, x, y, &x, &y);
+	    break;
+	case PLACE_MODE_CENTERED:
+	    placeCentered (window, &work_area, &x, &y);
+	    break;
+	case PLACE_MODE_RANDOM:
+	    placeRandom (window, &work_area, &x, &y);
+	    break;
+	case PLACE_MODE_SMART:
+	    placeSmart (window, &work_area, &x, &y);
+	    break;
+	case PLACE_MODE_MAXIMIZE:
+	    maximizeWindow (window, MAXIMIZE_STATE);
+	    break;
+	default:
+	    break;
+	}
+    }
 
     /* Maximize windows if they are too big for their work area (bit of
      * a hack here). Assume undecorated windows probably don't intend to
@@ -934,7 +1269,7 @@ placeWindow (CompWindow *window,
     {
 	XRectangle outer;
 
-	getOuterRectOfWindow (window, &outer);
+	get_outer_rect_of_window (window, &outer);
 
 	if (outer.width >= work_area.width && outer.height >= work_area.height)
 	    maximizeWindow (window, MAXIMIZE_STATE);
@@ -958,8 +1293,8 @@ done_check_denied_focus:
 	{
 	    XRectangle wr, fwr, overlap;
 
-	    getOuterRectOfWindow (window, &wr);
-	    getOuterRectOfWindow (focus_window, &fwr);
+	    get_outer_rect_of_window (window, &wr);
+	    get_outer_rect_of_window (focus_window, &fwr);
 
 	    /* No need to do anything if the window doesn't overlap at all */
 	    found_fit = !rectangleIntersect (&wr, &fwr, &overlap);
@@ -1015,34 +1350,37 @@ done_no_constraints:
 }
 
 static Bool
-placeDamageWindowRect (CompWindow *w,
-		       Bool	  initial,
-		       BoxPtr     rect)
+placePlaceWindow (CompWindow *w,
+		  int        x,
+		  int        y,
+		  int        *newX,
+		  int        *newY)
 {
     Bool status;
 
     PLACE_SCREEN (w->screen);
 
-    UNWRAP (ps, w->screen, damageWindowRect);
-    status = (*w->screen->damageWindowRect) (w, initial, rect);
-    WRAP (ps, w->screen, damageWindowRect, placeDamageWindowRect);
+    UNWRAP (ps, w->screen, placeWindow);
+    status = (*w->screen->placeWindow) (w, x, y, newX, newY);
+    WRAP (ps, w->screen, placeWindow, placePlaceWindow);
 
-    if (initial && !w->attrib.override_redirect && !w->placed)
+    if (!status)
     {
-	int newX, newY;
+	int viewportX, viewportY;
 
-	placeWindow (w, w->serverX, w->serverY, &newX, &newY);
+	placeWin (w, x, y, newX, newY);
 
-	w->placed = TRUE;
-
-	if (newX != w->serverX || newY != w->serverY)
+	if (placeMatchViewport (w, &viewportX, &viewportY))
 	{
-	    moveWindow (w, newX - w->attrib.x, newY - w->attrib.y, FALSE, TRUE);
-	    syncWindowPosition (w);
+	    viewportX = MAX (MIN (viewportX, w->screen->hsize), 0);
+	    viewportY = MAX (MIN (viewportY, w->screen->vsize), 0);
+
+	    *newX += (viewportX - w->screen->x) * w->screen->width;
+	    *newY += (viewportY - w->screen->y) * w->screen->height;
 	}
     }
 
-    return status;
+    return TRUE;
 }
 
 static Bool
@@ -1078,6 +1416,17 @@ placeFiniDisplay (CompPlugin  *p,
     free (pd);
 }
 
+static const CompMetadataOptionInfo placeScreenOptionInfo[] = {
+    { "workarounds", "bool", 0, 0, 0 },
+    { "mode", "int", RESTOSTRING (0, PLACE_MODE_LAST), 0, 0 },
+    { "position_matches", "list", "<type>match</type>", 0, 0 },
+    { "position_x_values", "list", "<type>int</type>", 0, 0 },
+    { "position_y_values", "list", "<type>int</type>", 0, 0 },
+    { "viewport_matches", "list", "<type>match</type>", 0, 0 },
+    { "viewport_x_values", "list", "<type>int</type>", 0, 0 },
+    { "viewport_y_values", "list", "<type>int</type>", 0, 0 }
+};
+
 static Bool
 placeInitScreen (CompPlugin *p,
 		 CompScreen *s)
@@ -1090,12 +1439,20 @@ placeInitScreen (CompPlugin *p,
     if (!ps)
 	return FALSE;
 
-    placeScreenInitOptions (ps);
-
-    WRAP (ps, s, damageWindowRect, placeDamageWindowRect);
+    if (!compInitScreenOptionsFromMetadata (s,
+					    &placeMetadata,
+					    placeScreenOptionInfo,
+					    ps->opt,
+					    PLACE_SCREEN_OPTION_NUM))
+    {
+	free (ps);
+	return FALSE;
+    }
 
     s->privates[pd->screenPrivateIndex].ptr = ps;
 
+    WRAP (ps, s, placeWindow, placePlaceWindow);
+
     return TRUE;
 }
 
@@ -1105,7 +1462,9 @@ placeFiniScreen (CompPlugin *p,
 {
     PLACE_SCREEN (s);
 
-    UNWRAP (ps, s, damageWindowRect);
+    UNWRAP (ps, s, placeWindow);
+
+    compFiniScreenOptions (s, ps->opt, PLACE_SCREEN_OPTION_NUM);
 
     free (ps);
 }
@@ -1113,9 +1472,20 @@ placeFiniScreen (CompPlugin *p,
 static Bool
 placeInit (CompPlugin *p)
 {
+    if (!compInitPluginMetadataFromInfo (&placeMetadata,
+					 p->vTable->name, 0, 0,
+					 placeScreenOptionInfo,
+					 PLACE_SCREEN_OPTION_NUM))
+	return FALSE;
+
     displayPrivateIndex = allocateDisplayPrivateIndex ();
     if (displayPrivateIndex < 0)
+    {
+	compFiniMetadata (&placeMetadata);
 	return FALSE;
+    }
+
+    compAddMetadataFromFile (&placeMetadata, p->vTable->name);
 
     return TRUE;
 }
@@ -1123,8 +1493,8 @@ placeInit (CompPlugin *p)
 static void
 placeFini (CompPlugin *p)
 {
-    if (displayPrivateIndex >= 0)
-	freeDisplayPrivateIndex (displayPrivateIndex);
+    freeDisplayPrivateIndex (displayPrivateIndex);
+    compFiniMetadata (&placeMetadata);
 }
 
 static int
@@ -1134,11 +1504,16 @@ placeGetVersion (CompPlugin *plugin,
     return ABIVERSION;
 }
 
+static CompMetadata *
+placeGetMetadata (CompPlugin *plugin)
+{
+    return &placeMetadata;
+}
+
 static CompPluginVTable placeVTable = {
     "place",
-    N_("Place Windows"),
-    N_("Place windows at appropriate positions when mapped"),
     placeGetVersion,
+    placeGetMetadata,
     placeInit,
     placeFini,
     placeInitDisplay,
@@ -1150,11 +1525,7 @@ static CompPluginVTable placeVTable = {
     0, /* GetDisplayOptions */
     0, /* SetDisplayOption */
     placeGetScreenOptions,
-    placeSetScreenOption,
-    0, /* Deps */
-    0, /* nDeps */
-    0, /* Features */
-    0  /* nFeatures */
+    placeSetScreenOption
 };
 
 CompPluginVTable *
