@@ -104,73 +104,11 @@ typedef struct _Model {
     unsigned int snapCnt[4];
 } Model;
 
-#define WOBBLY_FRICTION_DEFAULT    3.0f
-#define WOBBLY_FRICTION_MIN        0.1f
-#define WOBBLY_FRICTION_MAX       10.0f
-#define WOBBLY_FRICTION_PRECISION  0.1f
+#define WOBBLY_EFFECT_NONE   0
+#define WOBBLY_EFFECT_SHIVER 1
+#define WOBBLY_EFFECT_LAST   WOBBLY_EFFECT_SHIVER
 
-#define WOBBLY_SPRING_K_DEFAULT    8.0f
-#define WOBBLY_SPRING_K_MIN        0.1f
-#define WOBBLY_SPRING_K_MAX       10.0f
-#define WOBBLY_SPRING_K_PRECISION  0.1f
-
-#define WOBBLY_GRID_RESOLUTION_DEFAULT  8
-#define WOBBLY_GRID_RESOLUTION_MIN      1
-#define WOBBLY_GRID_RESOLUTION_MAX      64
-
-#define WOBBLY_MIN_GRID_SIZE_DEFAULT  8
-#define WOBBLY_MIN_GRID_SIZE_MIN      4
-#define WOBBLY_MIN_GRID_SIZE_MAX      128
-
-typedef enum {
-    WobblyEffectNone = 0,
-    WobblyEffectShiver
-} WobblyEffect;
-
-static char *effectName[] = {
-    N_("None"),
-    N_("Shiver")
-};
-
-static WobblyEffect effectType[] = {
-    WobblyEffectNone,
-    WobblyEffectShiver
-};
-
-#define NUM_EFFECT (sizeof (effectType) / sizeof (effectType[0]))
-
-#define WOBBLY_MAP_DEFAULT   (effectName[1])
-#define WOBBLY_FOCUS_DEFAULT (effectName[0])
-
-static char *mapWinType[] = {
-    N_("Splash"),
-    N_("DropdownMenu"),
-    N_("PopupMenu"),
-    N_("Tooltip"),
-    N_("Notification"),
-    N_("Combo"),
-    N_("Dnd"),
-    N_("Unknown")
-};
-#define N_MAP_WIN_TYPE (sizeof (mapWinType) / sizeof (mapWinType[0]))
-#define N_FOCUS_WIN_TYPE (0)
-
-static char *moveWinType[] = {
-    N_("Toolbar"),
-    N_("Menu"),
-    N_("Utility"),
-    N_("Dialog"),
-    N_("ModalDialog"),
-    N_("Normal")
-};
-#define N_MOVE_WIN_TYPE (sizeof (moveWinType) / sizeof (moveWinType[0]))
-#define N_GRAB_WIN_TYPE (0)
-
-#define WOBBLY_SNAP_MODIFIERS_DEFAULT ShiftMask
-
-#define WOBBLY_SNAP_INVERTED_DEFAULT FALSE
-
-#define WOBBLY_MAXIMIZE_EFFECT_DEFAULT TRUE
+static CompMetadata wobblyMetadata;
 
 static int displayPrivateIndex;
 
@@ -188,29 +126,28 @@ typedef struct _WobblyDisplay {
     Bool snapping;
 } WobblyDisplay;
 
-#define WOBBLY_SCREEN_OPTION_FRICTION	       0
-#define WOBBLY_SCREEN_OPTION_SPRING_K	       1
-#define WOBBLY_SCREEN_OPTION_GRID_RESOLUTION   2
-#define WOBBLY_SCREEN_OPTION_MIN_GRID_SIZE     3
-#define WOBBLY_SCREEN_OPTION_MAP_EFFECT	       4
-#define WOBBLY_SCREEN_OPTION_FOCUS_EFFECT      5
-#define WOBBLY_SCREEN_OPTION_MAP_WINDOW_TYPE   6
-#define WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_TYPE 7
-#define WOBBLY_SCREEN_OPTION_GRAB_WINDOW_TYPE  8
-#define WOBBLY_SCREEN_OPTION_MOVE_WINDOW_TYPE  9
-#define WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT   10
-#define WOBBLY_SCREEN_OPTION_NUM	       11
+#define WOBBLY_SCREEN_OPTION_FRICTION	        0
+#define WOBBLY_SCREEN_OPTION_SPRING_K	        1
+#define WOBBLY_SCREEN_OPTION_GRID_RESOLUTION    2
+#define WOBBLY_SCREEN_OPTION_MIN_GRID_SIZE      3
+#define WOBBLY_SCREEN_OPTION_MAP_EFFECT	        4
+#define WOBBLY_SCREEN_OPTION_FOCUS_EFFECT       5
+#define WOBBLY_SCREEN_OPTION_MAP_WINDOW_MATCH   6
+#define WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_MATCH 7
+#define WOBBLY_SCREEN_OPTION_GRAB_WINDOW_MATCH  8
+#define WOBBLY_SCREEN_OPTION_MOVE_WINDOW_MATCH  9
+#define WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT    10
+#define WOBBLY_SCREEN_OPTION_NUM	        11
 
 typedef struct _WobblyScreen {
     int	windowPrivateIndex;
 
     PreparePaintScreenProc preparePaintScreen;
     DonePaintScreenProc	   donePaintScreen;
-    PaintScreenProc	   paintScreen;
+    PaintOutputProc	   paintOutput;
     PaintWindowProc	   paintWindow;
     DamageWindowRectProc   damageWindowRect;
     AddWindowGeometryProc  addWindowGeometry;
-    DrawWindowGeometryProc drawWindowGeometry;
 
     WindowResizeNotifyProc windowResizeNotify;
     WindowMoveNotifyProc   windowMoveNotify;
@@ -221,16 +158,9 @@ typedef struct _WobblyScreen {
 
     Bool wobblyWindows;
 
-    WobblyEffect mapEffect;
-    WobblyEffect focusEffect;
-
-    unsigned int mapWMask;
-    unsigned int focusWMask;
-    unsigned int moveWMask;
-    unsigned int grabWMask;
-
     unsigned int grabMask;
     CompWindow	 *grabWindow;
+    Bool         moveWindow;
 } WobblyScreen;
 
 #define WobblyInitial  (1L << 0)
@@ -268,7 +198,8 @@ typedef struct _WobblyWindow {
 #define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))
 
 static CompOption *
-wobblyGetScreenOptions (CompScreen *screen,
+wobblyGetScreenOptions (CompPlugin *plugin,
+			CompScreen *screen,
 			int	   *count)
 {
     WOBBLY_SCREEN (screen);
@@ -278,228 +209,35 @@ wobblyGetScreenOptions (CompScreen *screen,
 }
 
 static Bool
-wobblySetScreenOption (CompScreen      *screen,
-		     char	     *name,
-		     CompOptionValue *value)
+wobblySetScreenOption (CompPlugin      *plugin,
+		       CompScreen      *screen,
+		       char	       *name,
+		       CompOptionValue *value)
 {
     CompOption *o;
-    int	       index;
 
     WOBBLY_SCREEN (screen);
 
-    o = compFindOption (ws->opt, NUM_OPTIONS (ws), name, &index);
+    o = compFindOption (ws->opt, NUM_OPTIONS (ws), name, NULL);
     if (!o)
 	return FALSE;
 
-    switch (index) {
-    case WOBBLY_SCREEN_OPTION_FRICTION:
-    case WOBBLY_SCREEN_OPTION_SPRING_K:
-	if (compSetFloatOption (o, value))
-	    return TRUE;
-	break;
-    case WOBBLY_SCREEN_OPTION_GRID_RESOLUTION:
-	if (compSetIntOption (o, value))
-	    return TRUE;
-	break;
-    case WOBBLY_SCREEN_OPTION_MIN_GRID_SIZE:
-	if (compSetIntOption (o, value))
-	    return TRUE;
-	break;
-    case WOBBLY_SCREEN_OPTION_MAP_EFFECT:
-	if (compSetStringOption (o, value))
-	{
-	    int i;
-
-	    for (i = 0; i < NUM_EFFECT; i++)
-	    {
-		if (strcmp (o->value.s, effectName[i]) == 0)
-		{
-		    ws->mapEffect = effectType[i];
-		    return TRUE;
-		}
-	    }
-	}
-	break;
-    case WOBBLY_SCREEN_OPTION_FOCUS_EFFECT:
-	if (compSetStringOption (o, value))
-	{
-	    int i;
-
-	    for (i = 0; i < NUM_EFFECT; i++)
-	    {
-		if (strcmp (o->value.s, effectName[i]) == 0)
-		{
-		    ws->focusEffect = effectType[i];
-		    return TRUE;
-		}
-	    }
-	}
-	break;
-    case WOBBLY_SCREEN_OPTION_MAP_WINDOW_TYPE:
-	if (compSetOptionList (o, value))
-	{
-	    ws->mapWMask = compWindowTypeMaskFromStringList (&o->value);
-	    return TRUE;
-	}
-	break;
-    case WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_TYPE:
-	if (compSetOptionList (o, value))
-	{
-	    ws->focusWMask = compWindowTypeMaskFromStringList (&o->value);
-	    return TRUE;
-	}
-	break;
-    case WOBBLY_SCREEN_OPTION_MOVE_WINDOW_TYPE:
-	if (compSetOptionList (o, value))
-	{
-	    ws->moveWMask = compWindowTypeMaskFromStringList (&o->value);
-	    return TRUE;
-	}
-	break;
-    case WOBBLY_SCREEN_OPTION_GRAB_WINDOW_TYPE:
-	if (compSetOptionList (o, value))
-	{
-	    ws->grabWMask = compWindowTypeMaskFromStringList (&o->value);
-	    return TRUE;
-	}
-	break;
-    case WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT:
-	if (compSetBoolOption (o, value))
-	    return TRUE;
-    default:
-	break;
-    }
-
-    return FALSE;
+    return compSetScreenOption (screen, o, value);
 }
 
-static void
-wobblyScreenInitOptions (WobblyScreen *ws,
-			 Display      *display)
-{
-    CompOption *o;
-    int	       i;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_FRICTION];
-    o->name		= "friction";
-    o->shortDesc	= N_("Friction");
-    o->longDesc		= N_("Spring Friction");
-    o->type		= CompOptionTypeFloat;
-    o->value.f		= WOBBLY_FRICTION_DEFAULT;
-    o->rest.f.min	= WOBBLY_FRICTION_MIN;
-    o->rest.f.max	= WOBBLY_FRICTION_MAX;
-    o->rest.f.precision = WOBBLY_FRICTION_PRECISION;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_SPRING_K];
-    o->name		= "spring_k";
-    o->shortDesc	= N_("Spring K");
-    o->longDesc		= N_("Spring Konstant");
-    o->type		= CompOptionTypeFloat;
-    o->value.f		= WOBBLY_SPRING_K_DEFAULT;
-    o->rest.f.min	= WOBBLY_SPRING_K_MIN;
-    o->rest.f.max	= WOBBLY_SPRING_K_MAX;
-    o->rest.f.precision = WOBBLY_SPRING_K_PRECISION;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_GRID_RESOLUTION];
-    o->name	  = "grid_resolution";
-    o->shortDesc  = N_("Grid Resolution");
-    o->longDesc	  = N_("Vertex Grid Resolution");
-    o->type	  = CompOptionTypeInt;
-    o->value.i	  = WOBBLY_GRID_RESOLUTION_DEFAULT;
-    o->rest.i.min = WOBBLY_GRID_RESOLUTION_MIN;
-    o->rest.i.max = WOBBLY_GRID_RESOLUTION_MAX;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_MIN_GRID_SIZE];
-    o->name	  = "min_grid_size";
-    o->shortDesc  = N_("Minimum Grid Size");
-    o->longDesc	  = N_("Minimum Vertex Grid Size");
-    o->type	  = CompOptionTypeInt;
-    o->value.i	  = WOBBLY_MIN_GRID_SIZE_DEFAULT;
-    o->rest.i.min = WOBBLY_MIN_GRID_SIZE_MIN;
-    o->rest.i.max = WOBBLY_MIN_GRID_SIZE_MAX;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_MAP_EFFECT];
-    o->name	      = "map_effect";
-    o->shortDesc      = N_("Map Effect");
-    o->longDesc	      = N_("Map Window Effect");
-    o->type	      = CompOptionTypeString;
-    o->value.s	      = strdup (WOBBLY_MAP_DEFAULT);
-    o->rest.s.string  = effectName;
-    o->rest.s.nString = NUM_EFFECT;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_FOCUS_EFFECT];
-    o->name	      = "focus_effect";
-    o->shortDesc      = N_("Focus Effect");
-    o->longDesc	      = N_("Focus Window Effect");
-    o->type	      = CompOptionTypeString;
-    o->value.s	      = strdup (WOBBLY_FOCUS_DEFAULT);
-    o->rest.s.string  = effectName;
-    o->rest.s.nString = NUM_EFFECT;
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_MAP_WINDOW_TYPE];
-    o->name	         = "map_window_types";
-    o->shortDesc         = N_("Map Window Types");
-    o->longDesc	         = N_("Window types that should wobble when mapped");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeString;
-    o->value.list.nValue = N_MAP_WIN_TYPE;
-    o->value.list.value  = malloc (sizeof (CompOptionValue) * N_MAP_WIN_TYPE);
-    for (i = 0; i < N_MAP_WIN_TYPE; i++)
-	o->value.list.value[i].s = strdup (mapWinType[i]);
-    o->rest.s.string     = windowTypeString;
-    o->rest.s.nString    = nWindowTypeString;
-
-    ws->mapWMask = compWindowTypeMaskFromStringList (&o->value);
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_TYPE];
-    o->name	         = "focus_window_types";
-    o->shortDesc         = N_("Focus Window Types");
-    o->longDesc	         = N_("Window types that should wobble when focused");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeString;
-    o->value.list.nValue = N_FOCUS_WIN_TYPE;
-    o->value.list.value  = NULL;
-    o->rest.s.string     = windowTypeString;
-    o->rest.s.nString    = nWindowTypeString;
-
-    ws->focusWMask = compWindowTypeMaskFromStringList (&o->value);
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_MOVE_WINDOW_TYPE];
-    o->name	         = "move_window_types";
-    o->shortDesc         = N_("Move Window Types");
-    o->longDesc	         = N_("Window types that should wobble when moved");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeString;
-    o->value.list.nValue = N_MOVE_WIN_TYPE;
-    o->value.list.value  = malloc (sizeof (CompOptionValue) * N_MOVE_WIN_TYPE);
-    for (i = 0; i < N_MOVE_WIN_TYPE; i++)
-	o->value.list.value[i].s = strdup (moveWinType[i]);
-    o->rest.s.string     = windowTypeString;
-    o->rest.s.nString    = nWindowTypeString;
-
-    ws->moveWMask = compWindowTypeMaskFromStringList (&o->value);
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_GRAB_WINDOW_TYPE];
-    o->name	         = "grab_window_types";
-    o->shortDesc         = N_("Grab Window Types");
-    o->longDesc	         = N_("Window types that should wobble when grabbed");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeString;
-    o->value.list.nValue = N_GRAB_WIN_TYPE;
-    o->value.list.value  = NULL;
-    o->rest.s.string     = windowTypeString;
-    o->rest.s.nString    = nWindowTypeString;
-
-    ws->grabWMask = compWindowTypeMaskFromStringList (&o->value);
-
-    o = &ws->opt[WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT];
-    o->name	  = "maximize_effect";
-    o->shortDesc  = N_("Maximize Effect");
-    o->longDesc	  = N_("Wobble effect when maximizing and unmaximizing "
-		       "windows");
-    o->type	  = CompOptionTypeBool;
-    o->value.b    = WOBBLY_MAXIMIZE_EFFECT_DEFAULT;
-}
+static const CompMetadataOptionInfo wobblyScreenOptionInfo[] = {
+    { "friction", "float", "<min>0.1</min>", 0, 0 },
+    { "spring_k", "float", "<min>0.1</min>", 0, 0 },
+    { "grid_resolution", "int", "<min>1</min><max>64</max>", 0, 0 },
+    { "min_grid_size", "int", "<min>4</min><max>128</max>", 0, 0 },
+    { "map_effect", "int", RESTOSTRING (0, WOBBLY_EFFECT_LAST), 0, 0 },
+    { "focus_effect", "int", RESTOSTRING (0, WOBBLY_EFFECT_LAST), 0, 0 },
+    { "map_window_match", "match", 0, 0, 0 },
+    { "focus_window_match", "match", 0, 0, 0 },
+    { "grab_window_match", "match", 0, 0, 0 },
+    { "move_window_match", "match", 0, 0, 0 },
+    { "maximize_effect", "bool", 0, 0, 0 }
+};
 
 #define SNAP_WINDOW_TYPE (CompWindowTypeNormalMask  | \
 			  CompWindowTypeToolbarMask | \
@@ -1027,6 +765,26 @@ modelSetMiddleAnchor (Model *model,
     model->anchorObject->immobile = TRUE;
 }
 
+static void
+modelSetTopAnchor (Model *model,
+		   int   x,
+		   int   y,
+		   int   width)
+{
+    float gx;
+
+    gx = ((GRID_WIDTH - 1) / 2 * width)  / (float) (GRID_WIDTH - 1);
+
+    if (model->anchorObject)
+	model->anchorObject->immobile = FALSE;
+
+    model->anchorObject = &model->objects[(GRID_WIDTH - 1) / 2];
+    model->anchorObject->position.x = x + gx;
+    model->anchorObject->position.y = y;
+
+    model->anchorObject->immobile = TRUE;
+}
+
 static void
 modelAddEdgeAnchors (Model *model,
 		     int   x,
@@ -1377,7 +1135,10 @@ createModel (int	  x,
     model->numObjects = GRID_WIDTH * GRID_HEIGHT;
     model->objects = malloc (sizeof (Object) * model->numObjects);
     if (!model->objects)
+    {
+	free (model);
 	return 0;
+    }
 
     model->anchorObject = 0;
     model->numSprings = 0;
@@ -1558,7 +1319,7 @@ modelStepObject (CompWindow *window,
 		    object->position.y > object->vertEdge.end)
 		    findNextWestEdge (window, object);
 
-		if (object->vertEdge.snapped == FALSE ||
+		if (!object->vertEdge.snapped ||
 		    objectReleaseWestEdge (window, model, object))
 		{
 		    object->position.x += object->velocity.x;
@@ -1593,7 +1354,7 @@ modelStepObject (CompWindow *window,
 		    object->position.y > object->vertEdge.end)
 		    findNextEastEdge (window, object);
 
-		if (object->vertEdge.snapped == FALSE ||
+		if (!object->vertEdge.snapped ||
 		    objectReleaseEastEdge (window, model, object))
 		{
 		    object->position.x += object->velocity.x;
@@ -1631,7 +1392,7 @@ modelStepObject (CompWindow *window,
 		    object->position.x > object->horzEdge.end)
 		    findNextNorthEdge (window, object);
 
-		if (object->horzEdge.snapped == FALSE ||
+		if (!object->horzEdge.snapped ||
 		    objectReleaseNorthEdge (window, model, object))
 		{
 		    object->position.y += object->velocity.y;
@@ -1666,7 +1427,7 @@ modelStepObject (CompWindow *window,
 		    object->position.x > object->horzEdge.end)
 		    findNextSouthEdge (window, object);
 
-		if (object->horzEdge.snapped == FALSE ||
+		if (!object->horzEdge.snapped ||
 		    objectReleaseSouthEdge (window, model, object))
 		{
 		    object->position.y += object->velocity.y;
@@ -1930,16 +1691,19 @@ wobblyPreparePaintScreen (CompScreen *s,
 		    {
 			ww->model = 0;
 
-			moveWindow (w,
-				    model->topLeft.x + w->output.left -
-				    w->attrib.x,
-				    model->topLeft.y + w->output.top -
-				    w->attrib.y,
-				    TRUE, TRUE);
+			if (w->attrib.x == w->serverX &&
+			    w->attrib.y == w->serverY)
+			{
+			    moveWindow (w,
+					model->topLeft.x + w->output.left -
+					w->attrib.x,
+					model->topLeft.y + w->output.top -
+					w->attrib.y,
+					TRUE, TRUE);
+			    syncWindowPosition (w);
+			}
 
 			ww->model = model;
-
-			syncWindowPosition (w);
 		    }
 
 		    if (!(s->damageMask & COMP_SCREEN_DAMAGE_ALL_MASK))
@@ -1995,6 +1759,46 @@ wobblyDonePaintScreen (CompScreen *s)
     WRAP (ws, s, donePaintScreen, wobblyDonePaintScreen);
 }
 
+static void
+wobblyDrawWindowGeometry (CompWindow *w)
+{
+    int     texUnit = w->texUnits;
+    int     currentTexUnit = 0;
+    int     stride = (1 + texUnit) * 2;
+    GLfloat *vertices = w->vertices + (stride - 2);
+
+    stride *= sizeof (GLfloat);
+
+    glVertexPointer (2, GL_FLOAT, stride, vertices);
+
+    while (texUnit--)
+    {
+	if (texUnit != currentTexUnit)
+	{
+	    w->screen->clientActiveTexture (GL_TEXTURE0_ARB + texUnit);
+	    glEnableClientState (GL_TEXTURE_COORD_ARRAY);
+	    currentTexUnit = texUnit;
+	}
+	vertices -= 2;
+	glTexCoordPointer (2, GL_FLOAT, stride, vertices);
+    }
+
+    glDrawElements (GL_QUADS, w->indexCount, GL_UNSIGNED_SHORT, w->indices);
+
+    /* disable all texture coordinate arrays except 0 */
+    texUnit = w->texUnits;
+    if (texUnit > 1)
+    {
+	while (--texUnit)
+	{
+	    (*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB + texUnit);
+	    glDisableClientState (GL_TEXTURE_COORD_ARRAY);
+	}
+
+	(*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB);
+    }
+}
+
 static void
 wobblyAddWindowGeometry (CompWindow *w,
 			 CompMatrix *matrix,
@@ -2049,7 +1853,7 @@ wobblyAddWindowGeometry (CompWindow *w,
 	vSize = 2 + nMatrix * 2;
 
 	nVertices = w->vCount;
-	nIndices  = w->vCount;
+	nIndices  = w->indexCount;
 
 	v = w->vertices + (nVertices * vSize);
 	i = w->indices  + nIndices;
@@ -2145,7 +1949,9 @@ wobblyAddWindowGeometry (CompWindow *w,
 	    pClip++;
 	}
 
-	w->vCount = nIndices;
+	w->vCount	      = nVertices;
+	w->indexCount	      = nIndices;
+	w->drawWindowGeometry = wobblyDrawWindowGeometry;
     }
     else
     {
@@ -2155,62 +1961,10 @@ wobblyAddWindowGeometry (CompWindow *w,
     }
 }
 
-static void
-wobblyDrawWindowGeometry (CompWindow *w)
-{
-    WOBBLY_WINDOW (w);
-
-    if (ww->wobbly)
-    {
-	int     texUnit = w->texUnits;
-	int     currentTexUnit = 0;
-	int     stride = (1 + texUnit) * 2;
-	GLfloat *vertices = w->vertices + (stride - 2);
-
-	stride *= sizeof (GLfloat);
-
-	glVertexPointer (2, GL_FLOAT, stride, vertices);
-
-	while (texUnit--)
-	{
-	    if (texUnit != currentTexUnit)
-	    {
-		w->screen->clientActiveTexture (GL_TEXTURE0_ARB + texUnit);
-		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
-		currentTexUnit = texUnit;
-	    }
-	    vertices -= 2;
-	    glTexCoordPointer (2, GL_FLOAT, stride, vertices);
-	}
-
-	glDrawElements (GL_QUADS, w->vCount, GL_UNSIGNED_SHORT, w->indices);
-
-	/* disable all texture coordinate arrays except 0 */
-	texUnit = w->texUnits;
-	if (texUnit > 1)
-	{
-	    while (--texUnit)
-	    {
-		(*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB + texUnit);
-		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
-	    }
-
-	    (*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB);
-	}
-    }
-    else
-    {
-	WOBBLY_SCREEN (w->screen);
-
-	UNWRAP (ws, w->screen, drawWindowGeometry);
-	(*w->screen->drawWindowGeometry) (w);
-	WRAP (ws, w->screen, drawWindowGeometry, wobblyDrawWindowGeometry);
-    }
-}
-
 static Bool
 wobblyPaintWindow (CompWindow		   *w,
 		   const WindowPaintAttrib *attrib,
+		   const CompTransform	   *transform,
 		   Region		   region,
 		   unsigned int		   mask)
 {
@@ -2220,15 +1974,10 @@ wobblyPaintWindow (CompWindow		   *w,
     WOBBLY_WINDOW (w);
 
     if (ww->wobbly)
-    {
-	if (mask & PAINT_WINDOW_SOLID_MASK)
-	    return FALSE;
-
 	mask |= PAINT_WINDOW_TRANSFORMED_MASK;
-    }
 
     UNWRAP (ws, w->screen, paintWindow);
-    status = (*w->screen->paintWindow) (w, attrib, region, mask);
+    status = (*w->screen->paintWindow) (w, attrib, transform, region, mask);
     WRAP (ws, w->screen, paintWindow, wobblyPaintWindow);
 
     return status;
@@ -2341,17 +2090,13 @@ static void
 wobblyHandleEvent (CompDisplay *d,
 		   XEvent      *event)
 {
-    Window     activeWindow = 0;
+    Window     activeWindow = d->activeWindow;
     CompWindow *w;
     CompScreen *s;
 
     WOBBLY_DISPLAY (d);
 
     switch (event->type) {
-    case PropertyNotify:
-	if (event->xproperty.atom == d->winActiveAtom)
-	    activeWindow = d->activeWindow;
-	break;
     case MapNotify:
 	w = findWindowAtDisplay (d, event->xmap.window);
 	if (w)
@@ -2416,8 +2161,8 @@ wobblyHandleEvent (CompDisplay *d,
 	{
 	    WOBBLY_SCREEN (s);
 
-	    if (ws->grabWindow			       &&
-		(ws->moveWMask & ws->grabWindow->type) &&
+	    if (ws->grabWindow &&
+		ws->moveWindow &&
 		ws->opt[WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT].value.b)
 	    {
 		WOBBLY_WINDOW (ws->grabWindow);
@@ -2451,44 +2196,46 @@ wobblyHandleEvent (CompDisplay *d,
 		}
 	    }
 	}
-	break;
-    case PropertyNotify:
-	if (event->xproperty.atom == d->winActiveAtom)
-	{
-	    if (d->activeWindow != activeWindow)
-	    {
-		w = findWindowAtDisplay (d, d->activeWindow);
-		if (w && isWobblyWin (w))
-		{
-		    WOBBLY_WINDOW (w);
-		    WOBBLY_SCREEN (w->screen);
-
-		    if ((ws->focusWMask & w->type) &&
-			ws->focusEffect		   &&
-			wobblyEnsureModel (w))
-		    {
-			switch (ws->focusEffect) {
-			case WobblyEffectShiver:
-			    modelAdjustObjectsForShiver (ww->model,
-							 WIN_X (w),
-							 WIN_Y (w),
-							 WIN_W (w),
-							 WIN_H (w));
-			default:
-			    break;
-			}
-
-			ww->wobbly |= WobblyInitial;
-			ws->wobblyWindows |= ww->wobbly;
-
-			damagePendingOnScreen (w->screen);
-		    }
-		}
-	    }
-	}
     default:
 	break;
     }
+
+    if (d->activeWindow != activeWindow)
+    {
+	w = findWindowAtDisplay (d, d->activeWindow);
+	if (w && isWobblyWin (w))
+	{
+	    int mIndex;
+	    int focusEffect;
+
+	    WOBBLY_WINDOW (w);
+	    WOBBLY_SCREEN (w->screen);
+
+	    mIndex      = WOBBLY_SCREEN_OPTION_FOCUS_WINDOW_MATCH;
+	    focusEffect = ws->opt[WOBBLY_SCREEN_OPTION_FOCUS_EFFECT].value.i;
+
+	    if ((focusEffect != WOBBLY_EFFECT_NONE)         &&
+		matchEval (&ws->opt[mIndex].value.match, w) &&
+		wobblyEnsureModel (w))
+	    {
+		switch (focusEffect) {
+		case WOBBLY_EFFECT_SHIVER:
+		    modelAdjustObjectsForShiver (ww->model,
+						 WIN_X (w),
+						 WIN_Y (w),
+						 WIN_W (w),
+						 WIN_H (w));
+		default:
+		    break;
+		}
+
+		ww->wobbly |= WobblyInitial;
+		ws->wobblyWindows |= ww->wobbly;
+
+		damagePendingOnScreen (w->screen);
+	    }
+	}
+    }
 }
 
 static Bool
@@ -2530,18 +2277,24 @@ wobblyDamageWindowRect (CompWindow *w,
     {
 	if (isWobblyWin (w))
 	{
+	    int mIndex;
+	    int mapEffect;
+
 	    WOBBLY_WINDOW (w);
 	    WOBBLY_SCREEN (w->screen);
 
+	    mIndex    = WOBBLY_SCREEN_OPTION_MAP_WINDOW_MATCH;
+	    mapEffect = ws->opt[WOBBLY_SCREEN_OPTION_MAP_EFFECT].value.i;
+
 	    if (ws->opt[WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT].value.b)
 		wobblyEnsureModel (w);
 
-	    if ((ws->mapWMask & w->type) &&
-		ws->mapEffect		 &&
+	    if ((mapEffect != WOBBLY_EFFECT_NONE)           &&
+		matchEval (&ws->opt[mIndex].value.match, w) &&
 		wobblyEnsureModel (w))
 	    {
-		switch (ws->mapEffect) {
-		case WobblyEffectShiver:
+		switch (mapEffect) {
+		case WOBBLY_EFFECT_SHIVER:
 		    modelAdjustObjectsForShiver (ww->model,
 						 WIN_X (w), WIN_Y (w),
 						 WIN_W (w), WIN_H (w));
@@ -2561,7 +2314,11 @@ wobblyDamageWindowRect (CompWindow *w,
 }
 
 static void
-wobblyWindowResizeNotify (CompWindow *w)
+wobblyWindowResizeNotify (CompWindow *w,
+			  int	     dx,
+			  int        dy,
+			  int	     dwidth,
+			  int	     dheight)
 {
     WOBBLY_SCREEN (w->screen);
     WOBBLY_WINDOW (w);
@@ -2608,9 +2365,16 @@ wobblyWindowResizeNotify (CompWindow *w)
     }
     else if (ww->model)
     {
-	if (!ww->wobbly)
+	if (ww->wobbly)
+	{
+	    if (!(ww->state & MAXIMIZE_STATE))
+		modelSetTopAnchor (ww->model, WIN_X (w), WIN_Y (w), WIN_W (w));
+	}
+	else
+	{
 	    modelInitObjects (ww->model,
 			      WIN_X (w), WIN_Y (w), WIN_W (w), WIN_H (w));
+	}
 
 	modelInitSprings (ww->model,
 			  WIN_X (w), WIN_Y (w), WIN_W (w), WIN_H (w));
@@ -2634,7 +2398,7 @@ wobblyWindowResizeNotify (CompWindow *w)
     }
 
     UNWRAP (ws, w->screen, windowResizeNotify);
-    (*w->screen->windowResizeNotify) (w);
+    (*w->screen->windowResizeNotify) (w, dx, dy, dwidth, dheight);
     WRAP (ws, w->screen, windowResizeNotify, wobblyWindowResizeNotify);
 }
 
@@ -2691,17 +2455,27 @@ wobblyWindowGrabNotify (CompWindow   *w,
 			unsigned int state,
 			unsigned int mask)
 {
+    int mIndex;
+
     WOBBLY_SCREEN (w->screen);
 
-    ws->grabMask   = mask;
-    ws->grabWindow = w;
+    mIndex = WOBBLY_SCREEN_OPTION_MOVE_WINDOW_MATCH;
 
-    if ((mask & CompWindowGrabButtonMask) &&
-	(ws->moveWMask & w->type)         &&
+    if (!ws->grabWindow)
+    {
+    	ws->grabMask   = mask;
+	ws->grabWindow = w;
+    }
+    ws->moveWindow = FALSE;
+
+    if ((mask & CompWindowGrabButtonMask)	    &&
+	matchEval (&ws->opt[mIndex].value.match, w) &&
 	isWobblyWin (w))
     {
 	WOBBLY_WINDOW (w);
 
+	ws->moveWindow = TRUE;
+
 	if (wobblyEnsureModel (w))
 	{
 	    Spring *s;
@@ -2745,7 +2519,9 @@ wobblyWindowGrabNotify (CompWindow   *w,
 		    modelUpdateSnapping (w, ww->model);
 	    }
 
-	    if (ws->grabWMask & w->type)
+	    mIndex = WOBBLY_SCREEN_OPTION_GRAB_WINDOW_MATCH;
+
+	    if (matchEval (&ws->opt[mIndex].value.match, w))
 	    {
 		for (i = 0; i < ww->model->numSprings; i++)
 		{
@@ -2782,8 +2558,11 @@ wobblyWindowUngrabNotify (CompWindow *w)
     WOBBLY_SCREEN (w->screen);
     WOBBLY_WINDOW (w);
 
-    ws->grabMask   = 0;
-    ws->grabWindow = NULL;
+    if (w == ws->grabWindow)
+    {
+    	ws->grabMask   = 0;
+	ws->grabWindow = NULL;
+    }
 
     if (ww->grabbed)
     {
@@ -2818,10 +2597,11 @@ wobblyWindowUngrabNotify (CompWindow *w)
 
 
 static Bool
-wobblyPaintScreen (CompScreen		   *s,
+wobblyPaintOutput (CompScreen		   *s,
 		   const ScreenPaintAttrib *sAttrib,
+		   const CompTransform	   *transform,
 		   Region		   region,
-		   int			   output,
+		   CompOutput		   *output,
 		   unsigned int		   mask)
 {
     Bool status;
@@ -2831,15 +2611,16 @@ wobblyPaintScreen (CompScreen		   *s,
     if (ws->wobblyWindows)
 	mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;
 
-    UNWRAP (ws, s, paintScreen);
-    status = (*s->paintScreen) (s, sAttrib, region, output, mask);
-    WRAP (ws, s, paintScreen, wobblyPaintScreen);
+    UNWRAP (ws, s, paintOutput);
+    status = (*s->paintOutput) (s, sAttrib, transform, region, output, mask);
+    WRAP (ws, s, paintOutput, wobblyPaintOutput);
 
     return status;
 }
 
 static CompOption *
-wobblyGetDisplayOptions (CompDisplay *display,
+wobblyGetDisplayOptions (CompPlugin  *plugin,
+			 CompDisplay *display,
 			 int	     *count)
 {
     WOBBLY_DISPLAY (display);
@@ -2849,7 +2630,8 @@ wobblyGetDisplayOptions (CompDisplay *display,
 }
 
 static Bool
-wobblySetDisplayOption (CompDisplay     *display,
+wobblySetDisplayOption (CompPlugin  *plugin,
+			CompDisplay     *display,
 			char	        *name,
 			CompOptionValue *value)
 {
@@ -2870,60 +2652,19 @@ wobblySetDisplayOption (CompDisplay     *display,
 	if (compSetActionOption (o, value))
 	    return TRUE;
 	break;
-    case WOBBLY_DISPLAY_OPTION_SNAP_INVERTED:
-	if (compSetBoolOption (o, value))
-	    return TRUE;
-	break;
-    case WOBBLY_DISPLAY_OPTION_SHIVER:
-	if (setDisplayAction (display, o, value))
-	    return TRUE;
     default:
-	break;
+	return compSetDisplayOption (display, o, value);
     }
 
     return FALSE;
 }
 
-static void
-wobblyDisplayInitOptions (WobblyDisplay *wd)
-{
-    CompOption *o;
-
-    o = &wd->opt[WOBBLY_DISPLAY_OPTION_SNAP];
-    o->name			  = "snap";
-    o->shortDesc		  = N_("Snap windows");
-    o->longDesc			  = N_("Toggle window snapping");
-    o->type			  = CompOptionTypeAction;
-    o->value.action.initiate	  = wobblyEnableSnapping;
-    o->value.action.terminate	  = wobblyDisableSnapping;
-    o->value.action.bell	  = FALSE;
-    o->value.action.edgeMask	  = 0;
-    o->value.action.state	  = 0;
-    o->value.action.type	  = CompBindingTypeKey;
-    o->value.action.key.modifiers = WOBBLY_SNAP_MODIFIERS_DEFAULT;
-    o->value.action.key.keycode   = 0;
-
-    o = &wd->opt[WOBBLY_DISPLAY_OPTION_SNAP_INVERTED];
-    o->name	 = "snap_inverted";
-    o->shortDesc = N_("Snap Inverted");
-    o->longDesc	 = N_("Inverted window snapping");
-    o->type	 = CompOptionTypeBool;
-    o->value.b	 = WOBBLY_SNAP_INVERTED_DEFAULT;
-
-    o = &wd->opt[WOBBLY_DISPLAY_OPTION_SHIVER];
-    o->name		      = "shiver";
-    o->shortDesc	      = N_("Shiver");
-    o->longDesc		      = N_("Make window shiver");
-    o->type		      = CompOptionTypeAction;
-    o->value.action.initiate  = wobblyShiver;
-    o->value.action.terminate = 0;
-    o->value.action.bell      = FALSE;
-    o->value.action.edgeMask  = 0;
-    o->value.action.state     = CompActionStateInitBell;
-    o->value.action.state    |= CompActionStateInitKey;
-    o->value.action.state    |= CompActionStateInitButton;
-    o->value.action.type      = 0;
-}
+static const CompMetadataOptionInfo wobblyDisplayOptionInfo[] = {
+    { "snap", "action", "<passive_grab>false</passive_grab>",
+      wobblyEnableSnapping, wobblyDisableSnapping },
+    { "snap_inverted", "bool", 0, 0, 0 },
+    { "shiver", "action", 0, wobblyShiver, 0 }
+};
 
 static Bool
 wobblyInitDisplay (CompPlugin  *p,
@@ -2935,9 +2676,20 @@ wobblyInitDisplay (CompPlugin  *p,
     if (!wd)
 	return FALSE;
 
+    if (!compInitDisplayOptionsFromMetadata (d,
+					     &wobblyMetadata,
+					     wobblyDisplayOptionInfo,
+					     wd->opt,
+					     WOBBLY_DISPLAY_OPTION_NUM))
+    {
+	free (wd);
+	return FALSE;
+    }
+
     wd->screenPrivateIndex = allocateScreenPrivateIndex (d);
     if (wd->screenPrivateIndex < 0)
     {
+	compFiniDisplayOptions (d, wd->opt, WOBBLY_DISPLAY_OPTION_NUM);
 	free (wd);
 	return FALSE;
     }
@@ -2946,8 +2698,6 @@ wobblyInitDisplay (CompPlugin  *p,
 
     wd->snapping = FALSE;
 
-    wobblyDisplayInitOptions (wd);
-
     d->privates[displayPrivateIndex].ptr = wd;
 
     return TRUE;
@@ -2963,6 +2713,8 @@ wobblyFiniDisplay (CompPlugin  *p,
 
     UNWRAP (wd, d, handleEvent);
 
+    compFiniDisplayOptions (d, wd->opt, WOBBLY_DISPLAY_OPTION_NUM);
+
     free (wd);
 }
 
@@ -2978,30 +2730,36 @@ wobblyInitScreen (CompPlugin *p,
     if (!ws)
 	return FALSE;
 
+    if (!compInitScreenOptionsFromMetadata (s,
+					    &wobblyMetadata,
+					    wobblyScreenOptionInfo,
+					    ws->opt,
+					    WOBBLY_SCREEN_OPTION_NUM))
+    {
+	free (ws);
+	return FALSE;
+    }
+
     ws->windowPrivateIndex = allocateWindowPrivateIndex (s);
     if (ws->windowPrivateIndex < 0)
     {
+	compFiniScreenOptions (s, ws->opt, WOBBLY_SCREEN_OPTION_NUM);
 	free (ws);
 	return FALSE;
     }
 
     ws->wobblyWindows = FALSE;
 
-    ws->mapEffect   = WobblyEffectShiver;
-    ws->focusEffect = WobblyEffectNone;
-
     ws->grabMask   = 0;
     ws->grabWindow = NULL;
-
-    wobblyScreenInitOptions (ws, s->display->display);
+    ws->moveWindow = FALSE;
 
     WRAP (ws, s, preparePaintScreen, wobblyPreparePaintScreen);
     WRAP (ws, s, donePaintScreen, wobblyDonePaintScreen);
-    WRAP (ws, s, paintScreen, wobblyPaintScreen);
+    WRAP (ws, s, paintOutput, wobblyPaintOutput);
     WRAP (ws, s, paintWindow, wobblyPaintWindow);
     WRAP (ws, s, damageWindowRect, wobblyDamageWindowRect);
     WRAP (ws, s, addWindowGeometry, wobblyAddWindowGeometry);
-    WRAP (ws, s, drawWindowGeometry, wobblyDrawWindowGeometry);
     WRAP (ws, s, windowResizeNotify, wobblyWindowResizeNotify);
     WRAP (ws, s, windowMoveNotify, wobblyWindowMoveNotify);
     WRAP (ws, s, windowGrabNotify, wobblyWindowGrabNotify);
@@ -3020,21 +2778,19 @@ wobblyFiniScreen (CompPlugin *p,
 
     freeWindowPrivateIndex (s, ws->windowPrivateIndex);
 
-    free (ws->opt[WOBBLY_SCREEN_OPTION_MAP_EFFECT].value.s);
-    free (ws->opt[WOBBLY_SCREEN_OPTION_FOCUS_EFFECT].value.s);
-
     UNWRAP (ws, s, preparePaintScreen);
     UNWRAP (ws, s, donePaintScreen);
-    UNWRAP (ws, s, paintScreen);
+    UNWRAP (ws, s, paintOutput);
     UNWRAP (ws, s, paintWindow);
     UNWRAP (ws, s, damageWindowRect);
     UNWRAP (ws, s, addWindowGeometry);
-    UNWRAP (ws, s, drawWindowGeometry);
     UNWRAP (ws, s, windowResizeNotify);
     UNWRAP (ws, s, windowMoveNotify);
     UNWRAP (ws, s, windowGrabNotify);
     UNWRAP (ws, s, windowUngrabNotify);
 
+    compFiniScreenOptions (s, ws->opt, WOBBLY_SCREEN_OPTION_NUM);
+
     free (ws);
 }
 
@@ -3084,9 +2840,22 @@ wobblyFiniWindow (CompPlugin *p,
 static Bool
 wobblyInit (CompPlugin *p)
 {
+    if (!compInitPluginMetadataFromInfo (&wobblyMetadata,
+					 p->vTable->name,
+					 wobblyDisplayOptionInfo,
+					 WOBBLY_DISPLAY_OPTION_NUM,
+					 wobblyScreenOptionInfo,
+					 WOBBLY_SCREEN_OPTION_NUM))
+	return FALSE;
+
     displayPrivateIndex = allocateDisplayPrivateIndex ();
     if (displayPrivateIndex < 0)
+    {
+	compFiniMetadata (&wobblyMetadata);
 	return FALSE;
+    }
+
+    compAddMetadataFromFile (&wobblyMetadata, p->vTable->name);
 
     return TRUE;
 }
@@ -3094,8 +2863,8 @@ wobblyInit (CompPlugin *p)
 static void
 wobblyFini (CompPlugin *p)
 {
-    if (displayPrivateIndex >= 0)
-	freeDisplayPrivateIndex (displayPrivateIndex);
+    freeDisplayPrivateIndex (displayPrivateIndex);
+    compFiniMetadata (&wobblyMetadata);
 }
 
 static int
@@ -3105,17 +2874,16 @@ wobblyGetVersion (CompPlugin *plugin,
     return ABIVERSION;
 }
 
-CompPluginDep wobblyDeps[] = {
-    { CompPluginRuleBefore, "fade" },
-    { CompPluginRuleBefore, "cube" },
-    { CompPluginRuleBefore, "scale" }
-};
+static CompMetadata *
+wobblyGetMetadata (CompPlugin *plugin)
+{
+    return &wobblyMetadata;
+}
 
 CompPluginVTable wobblyVTable = {
     "wobbly",
-    N_("Wobbly Windows"),
-    N_("Use spring model for wobbly window effect"),
     wobblyGetVersion,
+    wobblyGetMetadata,
     wobblyInit,
     wobblyFini,
     wobblyInitDisplay,
@@ -3127,11 +2895,7 @@ CompPluginVTable wobblyVTable = {
     wobblyGetDisplayOptions,
     wobblySetDisplayOption,
     wobblyGetScreenOptions,
-    wobblySetScreenOption,
-    wobblyDeps,
-    sizeof (wobblyDeps) / sizeof (wobblyDeps[0]),
-    0, /* Features */
-    0  /* nFeatures */
+    wobblySetScreenOption
 };
 
 CompPluginVTable *
